#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Sat May 28 22:04:03 2022
# @author: Ajit Johnson
# UMAP plot

"""
!!! abstract "Short Description"
    `sm.pl.umap`: The function allows users to generate a scatter plot of the UMAP. 
## Function
"""


# import lib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import math
import itertools
import matplotlib.colors as colors
import seaborn as sns
import matplotlib.patches as mpatches

# function
def umap (adata, color=None, use_layer=None, use_raw=False, log=False, label='umap',
          cmap='vlag', palette=None, alpha=0.8, figsize=(5, 5), s=None, ncols=None, 
          tight_layout=False, return_data=False, save_figure=None, **kwargs):
    """
Parameters:

    adata : AnnData Object  

    color : list, optional
        Keys for annotations of observations in `adata.obs.columns` or genes in `adata.var.index`. e.g. ['CD3D', 'SOX10']
        The default is None.
        
    use_layer : string, optional  
        Pass name of any `Layer` in AnnData. The default is `None` and `adata.X` is used.
        
    use_raw : bool, optional  
        If set to `True`, values in `adata.raw.X` will be used to color the plot. The default is False.
        
    log : bool, optional  
        If set to `True`, the data will natural log transformed using `np.log1p()` for coloring. The default is False.
        
    label : string, optional  
        The `label key` used when running `sm.tl.umap()`. The default is 'umap'.
        
    cmap : string, optional  
        Color map to use for continous variables. Can be a name or a Colormap 
        instance (e.g. "magma”, "viridis"). The default is 'vlag'.
        
    palette : dict, optional  
        Colors to use for plotting categorical annotation groups. 
        It accepts a `dict` mapping categories to colors. 
        e.g. `palette = {'T cells': '#000000', 'B cells': '#FFF675'}`.
        Auto color will be generated for other categories that are not specified. The default is None.
        
    alpha : float, optional  
        blending value, between 0 (transparent) and 1 (opaque). The default is 0.8.
        
    figsize : tuple, optional  
        Width, height in inches. The default is (10, 10).
        
    s : int, optional  
        The marker size in points. The default is None.
        
    ncols : int, optional  
        Number of panels per row. The default is None.
        
    tight_layout : bool, optional  
        Adjust the padding between and around subplots. If True it will ensure that
        the legends are visible. The default is False.
        
    return_data : bool, optional  
        Returns the data used for plotting. The default is False.
        
    save_figure : string, optional  
        Pass path to saving figure with file extension.
        e.g `\path\to\directory\figure.pdf` The default is None.
    
    **kwargs : Other `matplotlib` parameters. 

Returns:

    final_data : Dataframe
        If return_data is set to `True`.

Example:
```python

# Run UMAP
adata = sm.tl.umap(adata)
    
# plot results
sm.pl.umap(adata, color=['CD3D', 'SOX10'])


```

    """
    
    # check if umap tool has been run
    try:
        adata.obsm[label]
    except KeyError:
        raise KeyError("Please run `sm.tl.umap(adata)` first")
            
    # identify the coordinates
    umap_coordinates = pd.DataFrame(adata.obsm[label],index=adata.obs.index, columns=['umap-1','umap-2'])
    
    # other data that the user requests
    if color is not None:
        if isinstance(color, str):
            color = [color]
        # identify if all elemets of color are available        
        if set(color).issubset(list(adata.var.index) + list(adata.obs.columns)) is False:
            raise ValueError("Element passed to `color` is not found in adata, please check!")
        
        # organise the data
        if any(item in color for item in list(adata.obs.columns)):
            adataobs = adata.obs.loc[:, adata.obs.columns.isin(color)]
        else:
            adataobs = None
            
        if any(item in color for item in list(adata.var.index)):
            # find the index of the marker
            marker_index = np.where(np.isin(list(adata.var.index), color))[0]
            if use_layer is not None:
                adatavar = adata.layers[use_layer][:, np.r_[marker_index]]
            elif use_raw is True:
                adatavar = adata.raw.X[:, np.r_[marker_index]]
            else:
                adatavar = adata.X[:, np.r_[marker_index]]
            adatavar = pd.DataFrame(adatavar, index=adata.obs.index, columns = list(adata.var.index[marker_index]))
        else:
            adatavar = None

        # combine all color data
        if adataobs is not None and adatavar is not None:
            color_data = pd.concat ([adataobs, adatavar], axis=1)
        elif adataobs is not None and adatavar is None:
            color_data = adataobs
        elif adataobs is None and adatavar is not None:
            color_data = adatavar
    else:
        color_data = None
    
    # combine color data with umap coordinates
    if color_data is not None:
        final_data = pd.concat([umap_coordinates, color_data], axis=1)
    else:
        final_data = umap_coordinates
    
    # create some reasonable defaults
    # estimate number of columns in subpolt
    nplots = len(final_data.columns) - 2 # total number of plots
    if ncols is None:
        if nplots >= 4:
            subplot = [math.ceil(nplots / 4), 4]
        elif nplots == 0:
            subplot = [1, 1]
        else:
            subplot = [math.ceil(nplots / nplots), nplots]
    else:
        subplot = [math.ceil(nplots /ncols), ncols]
    
    if nplots == 0:
        n_plots_to_remove = 0
    else:
        n_plots_to_remove = np.prod(subplot) - nplots # figure if we have to remove any subplots
    
    # size of points
    if s is None:
        if nplots == 0:
            s = (100000 / adata.shape[0])
        else:
            s = (100000 / adata.shape[0]) / nplots
    

    # if there are categorical data then assign colors to them
    if final_data.select_dtypes(exclude=["number","bool_","object_"]).shape[1] > 0:
        # find all categories in the dataframe
        cat_data = final_data.select_dtypes(exclude=["number","bool_","object_"])
        # find all categories
        all_cat = []
        for i in cat_data.columns:
            all_cat.append(list(cat_data[i].cat.categories))
        
        # generate colormapping for all categories
        less_9 = [colors.rgb2hex(x) for x in sns.color_palette('Set1')]
        nineto20 = [colors.rgb2hex(x) for x in sns.color_palette('tab20')]
        greater20 = [colors.rgb2hex(x) for x in sns.color_palette('gist_ncar', max([len(i) for i in all_cat]))]
        
        all_cat_colormap = dict()
        for i in range(len(all_cat)):
            if len(all_cat[i]) <= 9:
                dict1 = dict(zip(all_cat[i] , less_9[ : len(all_cat[i]) ]   ))
            elif len(all_cat[i]) > 9 and len(all_cat[i]) <= 20:
                dict1 = dict(zip(all_cat[i] , nineto20[ : len(all_cat[i]) ]   ))
            else:
                dict1 = dict(zip(all_cat[i] , greater20[ : len(all_cat[i]) ]   ))
            all_cat_colormap.update(dict1)
        
        # if user has passed in custom colours update the colors
        if palette is not None:
            all_cat_colormap.update(palette)
    else:
        all_cat_colormap = None
        

    # plot
    fig, ax = plt.subplots(subplot[0],subplot[1], figsize=figsize)
    plt.rcdefaults()
    #plt.rcParams['axes.facecolor'] = 'white'
    
    # remove unwanted axes
    #fig.delaxes(ax[-1])
    if n_plots_to_remove > 0:
        for i in range(n_plots_to_remove):
            fig.delaxes(ax[-1][  (len(ax[-1])-1)-i :   (len(ax[-1]))-i   ][0])
    
    # to make sure the ax is always 2x2
    if any(i > 1 for i in subplot):
        if any(i == 1 for i in subplot):
            ax = ax.reshape(subplot[0],subplot[1])
    
    if nplots == 0:
        ax.scatter(x = final_data['umap-1'], y = final_data['umap-2'], s=s, cmap=cmap, alpha=alpha, **kwargs)
        plt.xlabel("UMAP-1"); plt.ylabel("UMAP-2") 
        plt.tick_params(right= False,top= False,left= False, bottom= False)
        ax.get_xaxis().set_ticks([]); ax.get_yaxis().set_ticks([])
        if tight_layout is True:
            plt.tight_layout()
    
    elif all(i == 1 for i in subplot):
        column_to_plot = [e for e in list(final_data.columns) if e not in ('umap-1', 'umap-2')][0]
        if all_cat_colormap is None:
            im = ax.scatter(x = final_data['umap-1'], y = final_data['umap-2'], s=s, 
                       c=final_data[column_to_plot],
                       cmap=cmap, alpha=alpha, **kwargs)
            plt.colorbar(im, ax=ax)
        else:
            ax.scatter(x = final_data['umap-1'], y = final_data['umap-2'], s=s, 
                       c=final_data[column_to_plot].map(all_cat_colormap),
                       cmap=cmap, alpha=alpha, **kwargs)
            # create legend
            patchList = []
            for key in list(final_data[column_to_plot].unique()):
                data_key = mpatches.Patch(color=all_cat_colormap[key], label=key)
                patchList.append(data_key)    
                ax.legend(handles=patchList,bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
        
        plt.xlabel("UMAP-1"); plt.ylabel("UMAP-2") 
        plt.title(column_to_plot)
        plt.tick_params(right= False,top= False,left= False, bottom= False)
        ax.set(xticklabels = ([])); ax.set(yticklabels = ([]))
        if tight_layout is True:
            plt.tight_layout()
        
    else: 
        column_to_plot = [e for e in list(final_data.columns) if e not in ('umap-1', 'umap-2')]
        k=0
        for i,j in itertools.product(range(subplot[0]), range(subplot[1])):
            
            if final_data[column_to_plot[k]].dtype == 'category':
                ax[i,j].scatter(x = final_data['umap-1'], y = final_data['umap-2'], s=s,
                                c=final_data[column_to_plot[k]].map(all_cat_colormap),
                                cmap=cmap, alpha=alpha, **kwargs)
                # create legend
                patchList = []
                for key in list(final_data[column_to_plot[k]].unique()):
                    data_key = mpatches.Patch(color=all_cat_colormap[key], label=key)
                    patchList.append(data_key)    
                    ax[i,j].legend(handles=patchList,bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
            else:               
                im = ax[i,j].scatter(x = final_data['umap-1'], y = final_data['umap-2'], s=s,
                                c=final_data[column_to_plot[k]],
                                cmap=cmap, alpha=alpha, **kwargs)
                plt.colorbar(im, ax=ax[i, j])

            ax[i,j].tick_params(right= False,top= False,left= False, bottom= False)
            ax[i,j].set_xticklabels([]); ax[i,j].set_yticklabels([])
            ax[i,j].set_xlabel("UMAP-1"); ax[i,j].set_ylabel("UMAP-2")
            ax[i,j].set_title(column_to_plot[k])
            if tight_layout is True:
                plt.tight_layout()
            k = k+1 # iterator
            
    # if save figure is requested
    if save_figure is not None:
        plt.savefig(save_figure)  
    
    # return data if needed
    if return_data is True:
        return final_data
            
    
    


from ._rename import rename
from ._classify import classify
from ._scimap_to_csv import scimap_to_csv
from ._addROI_omero import addROI_omero
from ._merge_adata_obs import merge_adata_obs
from ._animate import animate


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Sat Jan 30 21:38:21 2021
# @author: Ajit Johnson Nirmal

"""
!!! abstract "Short Description"
    `sm.pl.spatial_pscore`: The function allows users to plot proximity volume and density scores. 
    Run `sm.tl.spatial_pscore` before running this function. 

## Function
"""

# Library
import seaborn as sns; sns.set_theme(style='white', color_codes=True)
import matplotlib.pyplot as plt

# Function
def spatial_pscore (adata, label='spatial_pscore', plot_score='both', 
                    order_xaxis = None,
                    color='grey', **kwargs):
    """
Parameters:

    adata : AnnData object
    
    label : string, optional  
        The label under which the data is saved. This is the same `label` parameter 
        passed when running the `sm.tl.spatial_pscore` function.

    plot_score : string, optional  
        Three option are available. 
        A) Plot only the *Proximity Density* by passing in `Proximity Density`  
        B) Plot only the *Proximity Volume* by passing in `Proximity Volume`  
        C) Plot both side by side by passing `both`

    order_xaxis : list, optional  
        If the user wants to re-order the x-axis, pass all the names in the x-axis
        in the desired order as a list. e.g. ['ROI2', 'ROI4', "ROI1"] 

    color : string, optional  
        Color of the bars.

    **kwargs : string  
        Other arguments that can be passed into `sns.barplot`

Example:
```python
    # Plot only `Proximity Volume` scores
    sm.pl.spatial_pscore (adata, color='Black', 
    plot_score='Proximity Volume', order_xaxis=['ROI2', 'ROI4', "ROI1"])
```
    """
    
    
    # Isolate the data from anndata object
    data = adata.uns[label]
    
    # Order the x-axis if needed
    if order_xaxis is not None:
        data = data.reindex(order_xaxis)
        
     
    # Generate the x and y axis
    x  = data.index
    y_pd = data['Proximity Density'].values
    y_pv = data['Proximity Volume'].values
    
    # Plot what user requests
    if plot_score == 'Proximity Density':
        ax = sns.barplot(x, y_pd, color=color, **kwargs).set_title('Proximity Density')
        ax = plt.xticks(rotation=90)
        plt.tight_layout()
    if plot_score == 'Proximity Volume':
        ax = sns.barplot(x, y_pv, color=color, **kwargs).set_title('Proximity Volume')
        ax = plt.xticks(rotation=90)
        plt.tight_layout()
    if plot_score == 'both':
        fig, ax = plt.subplots(1,2)
        sns.barplot(x, y_pd, color=color, ax=ax[0], **kwargs).set_title('Proximity Density')
        ax[0].tick_params(axis='x', rotation=90)
        sns.barplot(x, y_pv, color=color, ax=ax[1], **kwargs).set_title('Proximity Volume')
        ax[1].tick_params(axis='x', rotation=90)
        plt.tight_layout()
        fig.show()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Fri Nov 27 09:34:22 2020
# @author: Ajit Johnson Nirmal

"""
!!! abstract "Short Description"
    `sm.pl.stacked_barplot`: The function allows users to generate a stacked bar plot of a categorical column. 
    The function can generate the plots using matplotlib and Plotly libraries. Plotly is browser based and so 
    it can be used for interactive data exploration.

## Function
"""

# Required library
import pandas as pd
import matplotlib.pylab as plt
import seaborn as sns; sns.set(color_codes=True)
import plotly.express as px
import plotly.graph_objects as go
import plotly.io as pio
pio.renderers.default = 'browser'
sns.set(style="white")


# Function
def stacked_barplot (adata, x_axis='imageid', y_axis='phenotype', subset_xaxis=None, subset_yaxis=None, 
                     order_xaxis=None, order_yaxis=None,
                     method='percent', plot_tool='matplotlib', matplotlib_cmap=None, 
                     matplotlib_bbox_to_anchor=(1,1.02), matplotlib_legend_loc=2, 
                     return_data=False, **kwargs):
    """
Parameters:
    adata : AnnData Object

    x_axis : string, required  
        Column name of the data that need to be plotted in the x-axis.

    y_axis : string, required  
        Column name of the data that need to be plotted in the y-axis.

    subset_xaxis : list, optional  
        Subset x-axis before plotting. Pass in a list of categories. eg- subset_xaxis = ['ROI_1', 'ROI_5']

    subset_yaxis : list, optional  
        Subset y-axis before plotting. Pass in a list of categories. eg- subset_yaxis = ['Celltype_A', 'Celltype_B']

    order_xaxis : list, optional  
        Order the x-axis of the plot as needed. Pass in a list of categories. eg- order_xaxis = ['ROI_5', 'ROI_1']
        The default is None and will be plotted based on alphabetic order. Please note that if you change the order, pass all categories, failure to do so
        will generate NaN's.

    order_yaxis : list, optional  
        Order the y-axis of the plot as needed. Pass in a list of categories. eg- order_yaxis = ['Celltype_B', 'Celltype_A']
        The default is None and will be plotted based on alphabetic order. Please note that if you change the order, pass all categories, failure to do so
        will generate NaN's.

    method : string, optional  
        Available options: 'percent' and 'absolute'. 
        1) Use Percent to plot the percent proportion.  
        2) Use 'absolute' to plot the plot the absolute number.  

    plot_tool : string, optional  
        Available options: 'matplotlib' and 'plotly'.  
        1) matplotlib uses the standard python plotting method  
        2) plotly opens the plot in a local browser. Advantage is to be able   
        to hover over the plot and retreive data for plots with large number of categories.

    matplotlib_cmap : string, optional  
        Colormap to select colors from. If string, load colormap with that name from matplotlib. 

    matplotlib_bbox_to_anchor : tuple, optional  
        Bounding box argument used along with matplotlib_legend_loc to control
        the legend location when using the matplotlib method.

    matplotlib_legend_loc : int, optional  
        Location of legend used along with matplotlib_bbox_to_anchor to control
        the legend location when using the matplotlib method.

    return_data : bool, optional  
        When True, return the data used for plotting.

    **kwargs : Additional keyword arguments passed to:  
        1) Pandas DataFrame.plot() when using the `matplotlib` method (https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.plot.html#pandas.DataFrame.plot))  
        2) Plotly.bar() when using the `plotly` method (https://plotly.com/python-api-reference/generated/plotly.express.bar.html))  

Returns:  
    Stacked bar plot. If return_data is set to `True` also returns a dataframe of the data used for the plot.

    
Example:
```python
    # Plot the absolute number of phenotypes using the matplotlib 
    # tool across differnt ROI's
    # ROI column is `epidermis_roi` and phenotypes are stored under `phenotype`

    sm.pl.stacked_barplot (adata,x_axis='epidermis_roi',y_axis='phenotype',
                     method='absolute',plot_tool='matplotlib',
                     figsize=(10, 10))
    
    # Plot the number of cells normalized to 100% using the plotly 
    # tool across differnt ROI's
    
    sm.pl.stacked_barplot (adata,x_axis='epidermis_roi',y_axis='phenotype',
                     method='percent',plot_tool='plotly',
                     color_discrete_sequence=px.colors.qualitative.Alphabet)
    
    # Same as above but visualize only a subset of ROI's and a subset of 
    # phenotypes
    subset_xaxis = ['epidermis_1', 'epidermis_5', 'epidermis_6']
    subset_yaxis = ['APCs', 'Keratinocytes', 'Macrophages']
    
    sm.pl.stacked_barplot (adata,x_axis='epidermis_roi',y_axis='phenotype',
                            subset_xaxis=subset_xaxis,subset_yaxis=subset_yaxis,
                            method='percent',plot_tool='plotly')
    
    # Visualize absolute number of phenotypes and return the data into a 
    # dataframe `absolute_number`
    absolute_number = sm.pl.stacked_barplot (adata,x_axis='epidermis_roi',
                      y_axis='phenotype', method='absolute',
                      plot_tool='matplotlib', return_data=True)

```
    """
    
    
    # create the dataframe with details
    data = pd.DataFrame(adata.obs)[[x_axis,y_axis]].astype(str)
    
    # subset the data if needed
    #if subset_data is not None:data = data[data[list(subset_data.keys())[0]].isin(list(subset_data.values())[0])]
    
    if subset_xaxis is not None:
        if isinstance(subset_xaxis, str):
            subset_xaxis = [subset_xaxis]
        data = data[data[x_axis].isin(subset_xaxis)]
    if subset_yaxis is not None:
        if isinstance(subset_yaxis, str):
            subset_yaxis = [subset_yaxis]
        data = data[data[y_axis].isin(subset_yaxis)]
        
        
    # Method: Absolute or Percentile
    if method == 'percent':
        total = data.groupby([x_axis,y_axis]).size().unstack().fillna(0).sum(axis=1)
        rg = pd.DataFrame(data.groupby([x_axis,y_axis]).size().unstack().fillna(0).div(total, axis=0).stack())
    elif method == 'absolute':
        rg = pd.DataFrame(data.groupby([x_axis,y_axis]).size().unstack().fillna(0).stack())
    else:
        raise ValueError('method should be either percent or absolute')
    
    # change column name
    rg.columns = ['count']
    
    # Add the index as columns in the data frame    
    rg.reset_index(inplace=True)  
    
    # re-order the x oy y axis if requested by user
    if order_xaxis is not None:
        rg[x_axis] = rg[x_axis].astype('category')
        rg[x_axis] = rg[x_axis].cat.reorder_categories(order_xaxis)
        rg = rg.sort_values(x_axis)
    if order_yaxis is not None:
        rg[y_axis] = rg[y_axis].astype('category')
        rg[y_axis] = rg[y_axis].cat.reorder_categories(order_yaxis)
        rg = rg.sort_values(y_axis)
    if order_xaxis and order_yaxis is not None:
        rg = rg.sort_values([x_axis, y_axis])
        
    pivot_df = rg.pivot(index=x_axis, columns=y_axis, values='count')
        
    # Plotting tool
    if plot_tool == 'matplotlib':
        
        if matplotlib_cmap is None:
            if len(rg[y_axis].unique()) <= 9:
                matplotlib_cmap = "Set1"        
            elif len(rg[y_axis].unique()) > 9 and len(rg[y_axis].unique()) <=20:
                matplotlib_cmap = plt.cm.tab20      #tab20  
            else:
                matplotlib_cmap = plt.cm.gist_ncar
        
        # Plotting
        # add width if not passed via parameters
        try:
            width
        except NameError:
            width=0.9
        # actual plotting   
        p = pivot_df.plot.bar(stacked=True, cmap=matplotlib_cmap, width=width,  **kwargs)
        handles, labels = p.get_legend_handles_labels() # for reversing the order of the legend
        p.legend(reversed(handles), reversed(labels), bbox_to_anchor=matplotlib_bbox_to_anchor, loc=matplotlib_legend_loc)
    
    elif plot_tool == 'plotly':
        
        fig = px.bar(rg, x=x_axis, y="count", color=y_axis, **kwargs)
        fig.update_layout({'plot_bgcolor': 'rgba(0, 0, 0, 0)',
                           'paper_bgcolor': 'rgba(0, 0, 0, 0)'},
                          xaxis = dict(tickmode='linear') #type = 'category'
                          )
        fig.show()
        
        
    else:
        
        raise ValueError('plot_tool should be either matplotlib or plotly')
    
    # Return data
    if return_data is True:
        return pivot_df
        

    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Thu Feb 11 20:19:50 2021
# @author: Ajit Johnson Nirmal
# Adapted from https://stackoverflow.com/questions/20515554/colorize-voronoi-diagram/20678647#20678647

"""
!!!
!!! abstract "Short Description"
    `sm.pl.voronoi`: The function allows users to generate a voronoi diagram and color it based on 
    any categorical column. Please note, voronoi diagrams are best fitted for small regions with 
    upto 5000 cells. Any larger, the voronoi plots are uninterpretable and takes a long time to generate.

## Function
"""

# import lib
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
try:
  from shapely.geometry import MultiPoint, Point, Polygon
except:
  print("If you are using Conda in Windows please install Shapely uisng the following command to use the Voronoi function\n\nconda install -c conda-forge shapely")
from scipy.spatial import Voronoi
import matplotlib.patches as mpatches

# Function
def voronoi_finite_polygons_2d(vor, radius=None):

    if vor.points.shape[1] != 2:
        raise ValueError("Requires 2D input")

    new_regions = []
    new_vertices = vor.vertices.tolist()

    center = vor.points.mean(axis=0)
    if radius is None:
        radius = vor.points.ptp().max()

    # Construct a map containing all ridges for a given point
    all_ridges = {}
    for (p1, p2), (v1, v2) in zip(vor.ridge_points, vor.ridge_vertices):
        all_ridges.setdefault(p1, []).append((p2, v1, v2))
        all_ridges.setdefault(p2, []).append((p1, v1, v2))

    # Reconstruct infinite regions
    for p1, region in enumerate(vor.point_region):
        vertices = vor.regions[region]

        if all(v >= 0 for v in vertices):
            # finite region
            new_regions.append(vertices)
            continue

        # reconstruct a non-finite region
        ridges = all_ridges[p1]
        new_region = [v for v in vertices if v >= 0]

        for p2, v1, v2 in ridges:
            if v2 < 0:
                v1, v2 = v2, v1
            if v1 >= 0:
                # finite ridge: already in the region
                continue

            # Compute the missing endpoint of an infinite ridge

            t = vor.points[p2] - vor.points[p1] # tangent
            t /= np.linalg.norm(t)
            n = np.array([-t[1], t[0]])  # normal

            midpoint = vor.points[[p1, p2]].mean(axis=0)
            direction = np.sign(np.dot(midpoint - center, n)) * n
            far_point = vor.vertices[v2] + direction * radius

            new_region.append(len(new_vertices))
            new_vertices.append(far_point.tolist())

        # sort region counterclockwise
        vs = np.asarray([new_vertices[v] for v in new_region])
        c = vs.mean(axis=0)
        angles = np.arctan2(vs[:,1] - c[1], vs[:,0] - c[0])
        new_region = np.array(new_region)[np.argsort(angles)]

        # finish
        new_regions.append(new_region.tolist())

    return new_regions, np.asarray(new_vertices)

# Actual function

def voronoi (adata, color_by=None, colors=None, x_coordinate='X_centroid', y_coordinate='Y_centroid',
             imageid='imageid',subset=None, x_lim=None, y_lim=None, flip_y=True,
             voronoi_edge_color='black', voronoi_line_width=0.1, voronoi_alpha=0.5, size_max=np.inf,
             overlay_points=None, overlay_points_categories=None, overlay_drop_categories=None, overlay_points_colors=None,
             overlay_point_size = 5, overlay_point_alpha= 1, overlay_point_shape=".", plot_legend=True, legend_size = 6, **kwargs):
    """
Parameters:
    adata : Anndata object

    color_by : string, optional  
        Color the voronoi diagram based on categorical variable (e.g. cell types or neighbourhoods).
        Pass the name of the column which contains the categorical variable.

    colors : string or Dict, optional  
        Custom coloring the voronoi diagram. The parameter accepts `sns color palettes` or a python dictionary
        mapping the categorical variable with the required color.

    x_coordinate : float, required  
        Column name containing the x-coordinates values.

    y_coordinate : float, required  
        Column name containing the y-coordinates values.
           
    flip_y : bool, optional  
        Flip the Y-axis if needed. Some algorithms output the XY with the Y-coordinates flipped.
        If the image overlays do not align to the cells, try again by setting this to `False`.

    imageid : string, optional  
        Column name of the column containing the image id.

    subset : string, optional  
        imageid of a single image to be subsetted for plotting.

    voronoi_edge_color : string, optional  
        A Matplotlib color for marking the edges of the voronoi. 
        If `facecolor` is passed, the edge color will always be the same as the face color.

    voronoi_line_width : float, optional  
        The linewidth of the marker edges. Note: The default edgecolors is 'face'. You may want to change this as well. 

    voronoi_alpha : float, optional  
        The alpha blending value, between 0 (transparent) and 1 (opaque).

    x_lim : list, optional  
        Pass the x-coordinates range [x1,x2].

    y_lim : list, optional  
        Pass the y-coordinates range [y1,y2].

    overlay_points : string, optional  
        It is possible to overlay a scatter plot on top of the voronoi diagram.
        Pass the name of the column which contains categorical variable to be overlayed.

    overlay_points_categories : list, optional  
        If the passed column in `overlay_points` contains multiple categories, however the user is only
        interested in a subset of categories, those specific names can be passed as a list. By default all 
        categories will be overlayed on the voronoi diagram.

    overlay_drop_categories : list, optional  
        Similar to `overlay_points_categories`. Here for ease of use, especially if large number of categories are present.
        The user can drop a set of categories.

    overlay_points_colors : string or dict, optional  
        Similar to `colors`.  
        User can pass in a  
        a) solid color (like `black`)  
        b) sns palettes name (like `Set1`)  
        c) python dictionary mapping the categories with custom colors

    overlay_point_size : float, optional  
        Overlay scatter plot point size.

    overlay_point_alpha : float, optional  
        The alpha blending value for the overlay, between 0 (transparent) and 1 (opaque).

    overlay_point_shape : string, optional  
        The marker style. marker can be either an instance of the class or the text shorthand for a particular marker.

    plot_legend : bool, optional  
        Define if the figure legend should be plotted.  
        Please note the figure legend may be out of view and you may need to resize the image to see it, especially 
        the legend for the scatter plot which will be on the left side of the plot.

    legend_size : float, optional  
        Resize the legend if needed.

Example:
    
    ```python
    sm.pl.voronoi(adata, color_by='phenotype', colors=None, 
             x_coordinate='X_position', y_coordinate='Y_position',
             imageid='ImageId',subset=None,
             voronoi_edge_color = 'black',voronoi_line_width = 0.2, 
             voronoi_alpha = 0.5, size_max=np.inf,
             overlay_points='phenotype', overlay_points_categories=None, 
             overlay_drop_categories=None,
             overlay_point_size = 5, overlay_point_alpha= 1, 
             overlay_point_shape=".", plot_legend=False, legend_size=6)
    
    ```

    """
    
    
    # create the data frame needed
    data = adata.obs
        
    # Subset the image of interest
    if subset is not None:
        data = data[data[imageid] == subset]
        

    # subset coordinates if needed
    if x_lim is not None:
        x1 = x_lim [0]
        if len(x_lim) < 2:
            x2 = max(data[x_coordinate])
        else:
            x2 = x_lim [1]
    if y_lim is not None:
        y1 = y_lim [0]
        if len(y_lim) < 2:
            y2 = min(data[y_coordinate])
        else:
            y2 = y_lim [1]
    # do the actuall subsetting
    if x_lim is not None:
        data = data[data[x_coordinate] >= x1]
        data = data[data[x_coordinate] <= x2]
    if y_lim is not None:
        data = data[data[y_coordinate] <= y1]
        data = data[data[y_coordinate] >= y2]

                    
    # create an extra column with index information
    data['index_info'] = np.arange(data.shape[0])
        
    # generate the x and y coordinates
    points = data[[x_coordinate,y_coordinate]].values
    
    # invert the Y-axis
    if flip_y is True:
        points[:,1] = max(points[:,1])-points[:,1]
    
    # Generate colors
    if color_by is None:
        colors = np.repeat('#e5e5e5', len(data))
#    elif color_by is None and colors is not None:
#        if isinstance(colors,str):
#            colors = np.repeat(colors, len(data))
    elif color_by is not None and colors is None:
        # auto color the samples
        if len(np.unique(data[color_by])) <= 9:
            c = sns.color_palette('Set1')[0:len(np.unique(data[color_by]))]
        if len(np.unique(data[color_by])) > 9 and len(np.unique(data[color_by])) <= 20:
            c = sns.color_palette('tab20')[0:len(np.unique(data[color_by]))]
        if len(np.unique(data[color_by])) > 20:
            # For large categories generate random colors 
            np.random.seed(0) ; c = np.random.rand(len(np.unique(data[color_by])),3).tolist()
        # merge colors with phenotypes/ categories of interest
        p = np.unique(data[color_by])
        c_p = dict(zip(p, c))
        # map to colors
        colors = list(map(c_p.get, list(data[color_by].values)))
    elif color_by is not None and colors is not None:
        # check if colors is a dictionary or a sns color scale
        if isinstance(colors,str): 
            if len(sns.color_palette(colors)) < len(np.unique(data[color_by])):
                raise ValueError(str(colors) + ' includes a maximun of ' + str(len(sns.color_palette(colors))) + ' colors, while your data need '  + str(len(np.unique(data[color_by]))) + ' colors')
            else:
                c = sns.color_palette(colors)[0:len(np.unique(data[color_by]))]    
                # merge colors with phenotypes/ categories of interest
                p = np.unique(data[color_by])
                c_p = dict(zip(p, c))
        if isinstance(colors,dict):
            if len(colors) < len(np.unique(data[color_by])):
                raise ValueError('Color mapping is not provided for all categories. Please check')
            else:
                c_p = colors
        # map to colors
        colors = list(map(c_p.get, list(data[color_by].values)))
        
        
    
    # create the voronoi object
    vor = Voronoi(points)
    
    # trim the object
    regions, vertices = voronoi_finite_polygons_2d(vor)
    
    # plotting
    pts = MultiPoint([Point(i) for i in points])
    mask = pts.convex_hull
    new_vertices = []
    if type(voronoi_alpha)!=list:
        voronoi_alpha = [voronoi_alpha]*len(points)
    areas = []
    for i,(region,alph) in enumerate(zip(regions,voronoi_alpha)):
        polygon = vertices[region]
        shape = list(polygon.shape)
        shape[0] += 1
        p = Polygon(np.append(polygon, polygon[0]).reshape(*shape)).intersection(mask)
        areas+=[p.area]
        if p.area <size_max:
            poly = np.array(list(zip(p.boundary.coords.xy[0][:-1], p.boundary.coords.xy[1][:-1])))
            new_vertices.append(poly)
            if voronoi_edge_color == 'facecolor':
                plt.fill(*zip(*poly), alpha=alph, edgecolor=colors[i], linewidth = voronoi_line_width , facecolor = colors[i])
                plt.xticks([]) ; plt.yticks([]);
            else:
                plt.fill(*zip(*poly), alpha=alph, edgecolor=voronoi_edge_color, linewidth = voronoi_line_width, facecolor = colors[i])
                plt.xticks([]) ; plt.yticks([]);
                #plt.xlim([1097.5,1414.5])
                #plt.ylim([167.3,464.1])

    
    # Add scatter on top of the voronoi if user requests
    if overlay_points is not None:
        if overlay_points_categories is None:
            d = data
        if overlay_points_categories is not None:
            # convert to list if needed (cells to keep)
            if isinstance(overlay_points_categories,str): 
                overlay_points_categories = [overlay_points_categories]
            # subset cells needed
            d = data[data[overlay_points].isin(overlay_points_categories)]    
        if overlay_drop_categories is not None:
            # conver to list if needed (cells to drop)
            if isinstance(overlay_drop_categories,str): 
                overlay_drop_categories = [overlay_drop_categories]
            # subset cells needed
            d = d[-d[overlay_points].isin(overlay_drop_categories)]
        
        # Find the x and y coordinates for the overlay category
        #points_scatter = d[[x_coordinate,y_coordinate]].values
        points_scatter = points[d.index_info.values]
    
        # invert the Y-axis
        #points_scatter[:,1] = max(points_scatter[:,1])-points_scatter[:,1]
        
        # Generate colors for the scatter plot
        if overlay_points_colors is None and color_by == overlay_points:
            # Borrow color from vornoi
            wanted_keys = np.unique(d[overlay_points]) # The keys to extract
            c_p_scatter = dict((k, c_p[k]) for k in wanted_keys if k in c_p)
        elif overlay_points_colors is None and color_by != overlay_points:
            # Randomly generate colors for all the categories in scatter plot
            # auto color the samples
            if len(np.unique(d[overlay_points])) <= 9:
                c_scatter = sns.color_palette('Set1')[0:len(np.unique(d[overlay_points]))]
            if len(np.unique(d[overlay_points])) > 9 and len(np.unique(d[overlay_points])) <= 20:
                c_scatter = sns.color_palette('tab20')[0:len(np.unique(d[overlay_points]))]
            if len(np.unique(d[overlay_points])) > 20:
                # For large categories generate random colors 
                np.random.seed(1) ; c_scatter = np.random.rand(len(np.unique(d[overlay_points])),3).tolist()
            # merge colors with phenotypes/ categories of interest
            p_scatter = np.unique(d[overlay_points])
            c_p_scatter = dict(zip(p_scatter, c_scatter))
        elif  overlay_points_colors is not None:
            # check if the overlay_points_colors is a pallete
            if isinstance(overlay_points_colors,str):
                try:
                    c_scatter = sns.color_palette(overlay_points_colors)[0:len(np.unique(d[overlay_points]))]
                    if len(sns.color_palette(overlay_points_colors)) < len(np.unique(d[overlay_points])):
                        raise ValueError(str(overlay_points_colors) + ' pallete includes a maximun of ' + str(len(sns.color_palette(overlay_points_colors))) + ' colors, while your data (overlay_points_colors) need '  + str(len(np.unique(d[overlay_points]))) + ' colors') 
                except:
                    c_scatter = np.repeat(overlay_points_colors,len(np.unique(d[overlay_points])))   #[overlay_points_colors]
                # create a dict
                p_scatter = np.unique(d[overlay_points])
                c_p_scatter = dict(zip(p_scatter, c_scatter))
            if isinstance(overlay_points_colors,dict):
                if len(overlay_points_colors) < len(np.unique(d[overlay_points])):
                    raise ValueError('Color mapping is not provided for all categories. Please check overlay_points_colors')
                else:
                    c_p_scatter = overlay_points_colors
        # map to colors
        colors_scatter = list(map(c_p_scatter.get, list(d[overlay_points].values)))
            
        #plt.scatter(x = points_scatter[:,0], y = points_scatter[:,1], s= overlay_point_size, alpha= overlay_point_alpha, c= colors_scatter, marker=overlay_point_shape)
        plt.scatter(x = points_scatter[:,0], y = points_scatter[:,1], s= overlay_point_size, alpha= overlay_point_alpha, c= colors_scatter, marker=overlay_point_shape,**kwargs)
        plt.xticks([]) ; plt.yticks([]);


    if plot_legend is True:
        # Add legend to voronoi
        patchList = []
        for key in c_p:
                data_key = mpatches.Patch(color=c_p[key], label=key)
                patchList.append(data_key)
    
        first_legend = plt.legend(handles=patchList, bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0., prop={'size': legend_size})
        plt.tight_layout()
        # Add the legend manually to the current Axes.
        ax = plt.gca().add_artist(first_legend)
        
        if overlay_points is not None:
            # Add legend to scatter
            patchList_scatter = []
            for key in c_p_scatter:
                    data_key_scatter = mpatches.Patch(color=c_p_scatter[key], label=key)
                    patchList_scatter.append(data_key_scatter)
        
            plt.legend(handles=patchList_scatter, bbox_to_anchor=(-0.05, 1), loc=1, borderaxespad=0., prop={'size': legend_size})
            #plt.tight_layout()
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Thu Jan 28 22:55:06 2021
# @author: Ajit Johnson Nirmal

"""
!!! abstract "Short Description"
    `sm.tl.spatial_pscore`: A scoring system to evaluate user defined proximity between cell types.  
    The function generates two scores and saved at `adata.uns`:   
    A) Proximity Density: Total number of interactions identified divided by the total number of 
    cells of the cell-types that were used for interaction analysis.  
    B) Proximity Volume: Total number of interactions identified divided by the total number of all cells in the data.  
    The interaction sites are also recorded and saved in `adata.obs`

## Functions
"""

# Import library
import pandas as pd
from sklearn.neighbors import BallTree
import numpy as np

# Function
def spatial_pscore (adata,proximity, score_by='imageid', x_coordinate='X_centroid',y_coordinate='Y_centroid',
                    phenotype='phenotype',method='radius',radius=20,knn=3,
                    imageid='imageid',subset=None, label='spatial_pscore'):
    """
Parameters:
    adata : AnnData object

    proximity : list  
        Pass a list of cell-types for which the proximity score needs to calculated. e.g. ['CellType-A', 'CellType-B']

    score_by : string, optional  
        If the scores need to compared across region's of interest, the column name containing the ROI's
        should be passed. By default the score is calculated across the entire image.

    x_coordinate : float, required  
        Column name containing the x-coordinates values.

    y_coordinate : float, required  
        Column name containing the y-coordinates values.

    phenotype : string, required  
        Column name of the column containing the phenotype information. 
        It could also be any categorical assignment given to single cells.

    method : string, optional  
        Two options are available: a) 'radius', b) 'knn'.  
        a) radius - Identifies the neighbours within a given radius for every cell.  
        b) knn - Identifies the K nearest neigbours for every cell.  

    radius : int, optional  
        The radius used to define a local neighbhourhood.

    knn : int, optional  
        Number of cells considered for defining the local neighbhourhood.

    imageid : string, optional  
        Column name of the column containing the image id.

    subset : string, optional  
        imageid of a single image to be subsetted for analyis.

    label : string, optional  
        Key for the returned data, stored in `adata.obs` and `adata.uns`.

Returns:
    adata : AnnData object  
        Updated AnnData object with the results stored in `adata.obs ['spatial_pscore']` and `adata.uns ['spatial_pscore']`.

Example:
```python
    # Calculate the score for proximity between `Tumor CD30+` cells and `M2 Macrophages`
    adata =  sm.tl.spatial_pscore (adata,proximity= ['Tumor CD30+', 'M2 Macrophages'], score_by = 'ImageId',
                             x_coordinate='X_position',y_coordinate='Y_position',
                             phenotype='phenotype',method='radius',radius=20,knn=3,
                             imageid='ImageId',subset=None, label='spatial_pscore')
    
```
    """
    
    
    # Start
    def spatial_pscore_internal (adata_subset,proximity,x_coordinate,y_coordinate,phenotype,method,radius,knn,
                                imageid,subset,label):

        # Create a DataFrame with the necessary inforamtion
        data = pd.DataFrame({'x': adata_subset.obs[x_coordinate], 'y': adata_subset.obs[y_coordinate], 'phenotype': adata_subset.obs[phenotype]})
        
        # Identify neighbourhoods based on the method used
        # a) KNN method
        if method == 'knn':
            print("Identifying the " + str(knn) + " nearest neighbours for every cell")
            tree = BallTree(data[['x','y']], leaf_size= 2)
            ind = tree.query(data[['x','y']], k=knn, return_distance= False)
            neighbours = pd.DataFrame(ind.tolist(), index = data.index) # neighbour DF
            neighbours_ind = neighbours.copy() # neighbour DF
            #neighbours.drop(0, axis=1, inplace=True) # Remove self neighbour
        
        # b) Local radius method
        if method == 'radius':
            print("Identifying neighbours within " + str(radius) + " pixels of every cell")
            kdt = BallTree(data[['x','y']], metric='euclidean') 
            ind = kdt.query_radius(data[['x','y']], r=radius, return_distance=False)
            #for i in range(0, len(ind)): ind[i] = np.delete(ind[i], np.argwhere(ind[i] == i))#remove self
            neighbours = pd.DataFrame(ind.tolist(), index = data.index) # neighbour DF
            neighbours_ind = neighbours.copy() # neighbour DF
            
        # Map phenotype
        phenomap = dict(zip(list(range(len(ind))), data['phenotype'])) # Used for mapping
        phenomap_ind = dict(zip(list(range(len(ind))), data.index)) # Used for mapping cell_nme
        
        # Loop through (all functionized methods were very slow)
        for i in neighbours.columns:
            neighbours[i] = neighbours[i].dropna().map(phenomap, na_action='ignore')
        # do the same index and cell name
        for i in neighbours_ind.columns:
            neighbours_ind[i] = neighbours_ind[i].dropna().map(phenomap_ind, na_action='ignore')
            
            
        # Idetify all the neighbourhoods that contains the user defined proximity phenotypes
        for i in proximity:
            print (str('Finding neighbourhoods with ') + str(i))
            nn = neighbours[neighbours.isin([i])].dropna(how='all').index
            neighbours = neighbours.loc[nn]
        
        # Identify all the cells that was part of the neighbourhood in this analysis
        neighbours_ind = neighbours_ind.loc[neighbours.index]
        neighbours_ind_unique = pd.unique(neighbours_ind.values.ravel())
        
        # subset the neighbourhood cells to include only the cells in the user defined list
        cleaned_neighbours_ind_unique = [x for x in neighbours_ind_unique if str(x) != 'nan']
        d = data.loc[cleaned_neighbours_ind_unique]
        d = d[d['phenotype'].isin(proximity)].index
        
        # return neighbours for score and image_neighbours for plotting on image
        return {'neighbours': neighbours.index, 'image_neighbours': d }
        
        
    # Subset a particular image if needed
    if subset is not None:
        adata_list = [adata[adata.obs[imageid] == subset]]
    else:
        adata_list = [adata[adata.obs[imageid] == i] for i in adata.obs[imageid].unique()]
    
    # Apply function to all images and create a master dataframe
    # Create lamda function 
    r_spatial_pscore_internal = lambda x: spatial_pscore_internal(adata_subset=x,proximity=proximity,
                                                   x_coordinate=x_coordinate,
                                                   y_coordinate=y_coordinate,phenotype=phenotype,
                                                   method=method,radius=radius,knn=knn,
                                                   imageid=imageid,subset=subset,label=label) 
    all_data = list(map(r_spatial_pscore_internal, adata_list)) # Apply function 
    
    
    # Merge all the results into a single dataframe    
    proximity_site_cells =  np.concatenate([d['image_neighbours'] for d in all_data], axis=0)
    
    # Add it to the AnnData Object
    adata.obs[label] = np.where(adata.obs.index.isin(proximity_site_cells), '_'.join(proximity), "other")
    
    
    ##### SCORING #####
    proximity_neigh = np.concatenate([d['neighbours'] for d in all_data], axis=0)
    wh_d = adata.obs
    wh_d[label] = np.where(wh_d.index.isin(proximity_neigh), '_'.join(proximity), "other")
    
    # Define a scoring system
    name = '_'.join(proximity)
    whole_data = wh_d[[score_by, label, phenotype]]
    
    # proximity volume
    p_v = whole_data.groupby([score_by, label]).size().unstack().fillna(0)
    p_v ['All Cells'] = p_v[name] + p_v["other"]
    p_v['Proximity Volume'] = p_v[name] / p_v['All Cells']
    p_v = p_v.fillna(0) # replace NA
    p_v = p_v.replace([np.inf, -np.inf], 0)
    p_v = p_v.drop(columns = 'other')
    
    # subset the phenotypes of interest
    w_d = whole_data[whole_data[phenotype].isin(proximity)]
    # proximity density
    p_d = w_d.groupby([score_by, label]).size().unstack().fillna(0)
    p_d ['Celltype of interest'] = p_d[name] + p_d["other"]
    p_d['Proximity Density'] = p_d[name] / p_d['Celltype of interest']
    p_d = p_d.fillna(0) # replace NA
    p_d = p_d.replace([np.inf, -np.inf], 0)
    p_d = p_d.drop(columns = ['other', name])
    
    # Merge Promimity volumne and density
    proximity_score = pd.merge(p_v, p_d, left_index=True, right_index=True)
    
    # Add it to the anndata object
    adata.uns[label] = proximity_score
    
    # Print
    print("Please check:\nadata.obs['" + str(label) + "'] &\nadata.uns['"+ str(label) + "'] for results")
      
    # Return 
    return adata

         


from ._phenotype_cells import phenotype_cells
from ._spatial_count import spatial_count
from ._spatial_expression import spatial_expression
from ._spatial_aggregate import spatial_aggregate
from ._cluster import cluster 
from ._spatial_interaction import spatial_interaction 
from ._spatial_distance import spatial_distance 
from ._spatial_pscore import spatial_pscore 
from ._spatial_lda import spatial_lda 
from ._foldchange import foldchange 
from ._spatial_cluster import spatial_cluster
from ._spatial_similarity_search import spatial_similarity_search
from ._umap import umap


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Wed Aug 19 15:00:39 2020
# @author: Ajit Johnson Nirmal
"""
!!! abstract "Short Description"
    The `sm.tl.spatial_count` function allows users to compute a neighbourhood matrix 
    using any categorical variable (e.g. cell-types) as input.

    The function supports two methods to define a local neighbourhood <br>
    **Radius method**: Can be used to identifies the neighbours within a user defined radius for every cell.
    **KNN method**: Can be used to identifies the neighbours based on K nearest neigbours for every cell
        
    The resultant neighbourhood matrix is saved with `adata.uns`. 

    This can be further clustered to identify similar neighbourhoods. 
    Use the [spatial_cluster] function to further group the neighbourhoods into 
    Reccurent Cellular Neighbourhoods (RCNs)

## Function
"""

# Import library
import pandas as pd
import numpy as np
from sklearn.neighbors import BallTree

# Function
def spatial_count (adata,
                   x_coordinate='X_centroid',
                   y_coordinate='Y_centroid',
                   phenotype='phenotype',
                   method='radius',
                   radius=30,knn=10,
                   imageid='imageid',
                   subset=None,
                   label='spatial_count'):
    """
Parameters:
    adata : anndata object

    x_coordinate : float, required  
        Column name containing the x-coordinates values.
        
    y_coordinate : float, required  
        Column name containing the y-coordinates values.
        
    phenotype : string, required  
        Column name of the column containing the phenotype information. 
        It could also be any categorical assignment given to single cells.
        
    method : string, optional  
        Two options are available: a) `radius`, b) `knn`.  
        a) radius - Identifies the neighbours within a given radius for every cell.  
        b) knn - Identifies the K nearest neigbours for every cell.  
        
    radius : int, optional  
        The radius used to define a local neighbhourhood.
        
    knn : int, optional  
        Number of cells considered for defining the local neighbhourhood.
        
    imageid : string, optional  
        Column name of the column containing the image id.
        
    subset : string, optional  
        imageid of a single image to be subsetted for analyis.
        
    label : string, optional  
        Key for the returned data, stored in `adata.uns`.

Returns:
    adata : anndata object  
        Updated AnnData object with the results stored in `adata.uns ['spatial_count']`.
    
Example:
    ```python
    # Running the radius method
    adata = sm.tl.spatial_count (adata,x_coordinate='X_centroid',
                                 y_coordinate='Y_centroid',
                                 phenotype='phenotype',
                                 method='radius',radius=30,
                                 imageid='imageid',subset=None,
                                 label='spatial_count_radius')
    
    # Running the knn method
    adata = sm.tl.spatial_count (adata,x_coordinate='X_centroid',
                                 y_coordinate='Y_centroid',
                                 phenotype='phenotype',method='knn',
                                 knn=10, imageid='imageid',
                                 subset=None,label='spatial_count_knn')
    ```
    """

    def spatial_count_internal (adata_subset,x_coordinate,y_coordinate,phenotype,method,radius,knn,
                                imageid,subset,label):

        # Create a DataFrame with the necessary inforamtion
        data = pd.DataFrame({'x': adata_subset.obs[x_coordinate], 'y': adata_subset.obs[y_coordinate], 'phenotype': adata_subset.obs[phenotype]})
        
        # Identify neighbourhoods based on the method used
        # a) KNN method
        if method == 'knn':
            print("Identifying the " + str(knn) + " nearest neighbours for every cell")
            tree = BallTree(data[['x','y']], leaf_size= 2)
            ind = tree.query(data[['x','y']], k=knn, return_distance= False)
            neighbours = pd.DataFrame(ind.tolist(), index = data.index) # neighbour DF
            neighbours.drop(0, axis=1, inplace=True) # Remove self neighbour
        
        # b) Local radius method
        if method == 'radius':
            print("Identifying neighbours within " + str(radius) + " pixels of every cell")
            kdt = BallTree(data[['x','y']], metric='euclidean') 
            ind = kdt.query_radius(data[['x','y']], r=radius, return_distance=False)
            for i in range(0, len(ind)): ind[i] = np.delete(ind[i], np.argwhere(ind[i] == i))#remove self
            neighbours = pd.DataFrame(ind.tolist(), index = data.index) # neighbour DF
            
        # Map phenotype
        phenomap = dict(zip(list(range(len(ind))), data['phenotype'])) # Used for mapping
        
        # Loop through (all functionized methods were very slow)
        for i in neighbours.columns:
            neighbours[i] = neighbours[i].dropna().map(phenomap, na_action='ignore')
        
        # Drop NA
        #n_dropped = neighbours.dropna(how='all')
           
        # Collapse all the neighbours into a single column
        n = pd.DataFrame(neighbours.stack(), columns = ["neighbour_phenotype"])
        n.index = n.index.get_level_values(0) # Drop the multi index
        n = pd.DataFrame(n)
        n['order'] = list(range(len(n)))
        
        # Merge with real phenotype
        n_m = n.merge(data['phenotype'], how='inner', left_index=True, right_index=True)
        n_m['neighbourhood'] = n_m.index
        n = n_m.sort_values(by=['order'])
        
        # Normalize based on total cell count
        k = n.groupby(['neighbourhood','neighbour_phenotype']).size().unstack().fillna(0)
        k = k.div(k.sum(axis=1), axis=0)
        
        # return the normalized neighbour occurance count
        return k
    
    # Subset a particular image if needed
    if subset is not None:
        adata_list = [adata[adata.obs[imageid] == subset]]
    else:
        adata_list = [adata[adata.obs[imageid] == i] for i in adata.obs[imageid].unique()]
    
    # Apply function to all images and create a master dataframe
    # Create lamda function 
    r_spatial_count_internal = lambda x: spatial_count_internal(adata_subset=x,x_coordinate=x_coordinate,
                                                   y_coordinate=y_coordinate,phenotype=phenotype,
                                                   method=method,radius=radius,knn=knn,
                                                   imageid=imageid,subset=subset,label=label) 
    all_data = list(map(r_spatial_count_internal, adata_list)) # Apply function 
    
    
    # Merge all the results into a single dataframe    
    result = []
    for i in range(len(all_data)):
        result.append(all_data[i])
    result = pd.concat(result, join='outer')  
    
    # Reindex the cells
    result = result.fillna(0)
    result = result.reindex(adata.obs.index)
    
    # Add to adata
    adata.uns[label] = result
    
    # Return        
    return adata


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Wed Apr  1 21:57:54 2020
# @author: Ajit Johnson Nirmal, Yuan Chen
"""
!!! abstract "Short Description"
    `sm.pl.image_viewer`: The function allows users to open OME-TIFF images inside 
    Napari and overlay any any categorical column such as cluster annotation or phenotypes.

## Function
"""

#%gui qt
try:
    import napari
except:
    pass
import pandas as pd
import random
import tifffile as tiff

import dask.array as da
from dask.cache import Cache
import zarr
import os

cache = Cache(2e9)  # Leverage two gigabytes of memory
cache.register()

def image_viewer (image_path, adata, overlay=None, flip_y=True,
                    overlay_category=None,markers=None,channel_names='default',
                    x_coordinate='X_centroid',y_coordinate='Y_centroid',point_size=10,
                    point_color=None,subset=None,imageid='imageid',seg_mask=None,**kwargs):
    """
Parameters:
    image_path : string  
        Location to the image file (TIFF, OME.TIFF, ZARR supported)

    seg_mask: string  
        Location to the segmentation mask file.

    adata : AnnData Object  
        
    flip_y : bool, optional  
        Flip the Y-axis if needed. Some algorithms output the XY with the Y-coordinates flipped.
        If the image overlays do not align to the cells, try again by setting this to `False`.

    overlay : string, optional  
        Name of the column with any categorical data such as phenotypes or clusters.

    overlay_category : list, optional  
        If only specfic categories within the overlay column is needed, pass their names as a list.
        If None, all categories will be used.

    markers : list, optional  
        Markers to be included. If none, all markers will be displayed.

    channel_names : list, optional  
        List of channels in the image in the exact order as image. The default is `adata.uns['all_markers']`

    x_coordinate : string, optional  
        X axis coordinate column name in AnnData object.

    y_coordinate : string, optional  
        Y axis coordinate column name in AnnData object.

    point_size : int, optional  
        point size in the napari plot.

    imageid : string, optional  
        Column name of the column containing the image id. 

    subset : string, optional  
        imageid of a single image to be subsetted for analyis. Only useful when multiple images are being analyzed together.

    **kwargs  
        Other arguments that can be passed to napari viewer
    
Returns:
    Napari Viewer.

Example:
```python
    image_path = '/Users/aj/Desktop/ptcl_tma/image.ome.tif'
    sm.pl.image_viewer (image_path, adata, overlay='phenotype',overlay_category=None,
                markers=['CD31', "CD3D","DNA11",'CD19','CD45','CD163','FOXP3'],
                point_size=7,point_color='white')
```
    """
    
    #TODO
    # - ADD Subset markers for ZARR ssection
    # - Ability to use ZARR metadata if available
    
    # adding option to load just the image without an adata object
    if adata is None:
        channel_names = None
    else: 
        # All operations on the AnnData object is performed first
        # Plot only the Image that is requested
        if subset is not None:
            adata = adata[adata.obs[imageid] == subset]
    
        # Recover the channel names from adata
        if channel_names == 'default':
            channel_names = adata.uns['all_markers']
        else:
            channel_names = channel_names
    
        # Index of the marker of interest and corresponding names
        if markers is None:
            idx = list(range(len(channel_names)))
            channel_names = channel_names
        else:
            idx = []
            for i in markers:
                idx.append(list(channel_names).index(i))
            channel_names = markers
        
        # Load the segmentation mask
        if seg_mask is not None:
            seg_m = tiff.imread(seg_mask)
            if (len(seg_m.shape) > 2) and (seg_m.shape[0] > 1):
                seg_m = seg_m[0]
    

    # Operations on the OME TIFF image is performed next
    # check the format of image
    if os.path.isfile(image_path) is True:  
        image = tiff.TiffFile(image_path, is_ome=False) #is_ome=False
        z = zarr.open(image.aszarr(), mode='r') # convert image to Zarr array
        # Identify the number of pyramids
        n_levels = len(image.series[0].levels) # pyramid
        
        # If and if not pyramids are available
        if n_levels > 1:
            pyramid = [da.from_zarr(z[i]) for i in range(n_levels)]
            multiscale = True
        else:
            pyramid = da.from_zarr(z)
            multiscale = False  
   
        # subset channels of interest
        if markers is not None:
            if n_levels > 1:
                for i in range(n_levels-1):
                    pyramid[i] = pyramid[i][idx, :, :]
                n_channels = pyramid[0].shape[0] # identify the number of channels
            else:
                pyramid = pyramid[idx, :, :]
                n_channels = pyramid.shape[0] # identify the number of channels
        else:
            if n_levels > 1:
                n_channels = pyramid[0].shape[0]
            else:
                n_channels = pyramid.shape[0]
            
        # check if channel names have been passed to all channels
        if channel_names is not None:
            assert n_channels == len(channel_names), (
                f'number of channel names ({len(channel_names)}) must '
                f'match number of channels ({n_channels})'
            )

        # Load the viewer
        viewer = napari.view_image(
            pyramid, multiscale=multiscale, channel_axis=0,
            visible=False, 
            name = None if channel_names is None else channel_names,
            #**kwargs
        )
    
    # Operations on the ZARR image
    # check the format of image
    if os.path.isfile(image_path) is False: 
        #print(image_path)
        viewer = napari.Viewer()
        viewer.open(image_path, multiscale=True,
                    visible=False,
                    name = None if channel_names is None else channel_names)
    
    # Add the seg mask
    if seg_mask is not None:
        viewer.add_labels(seg_m, name='segmentation mask', visible=False)

    # Add phenotype layer function
    def add_phenotype_layer (adata, overlay, phenotype_layer,x,y,viewer,point_size,point_color):
        coordinates = adata[adata.obs[overlay] == phenotype_layer]
        # Flip Y AXIS if needed
        if flip_y is True:
            coordinates = pd.DataFrame({'y': coordinates.obs[y],'x': coordinates.obs[x]})
        else:  
            coordinates = pd.DataFrame({'x': coordinates.obs[x],'y': coordinates.obs[y]})

        #points = coordinates.values.tolist()
        points = coordinates.values
        if point_color is None:
            r = lambda: random.randint(0,255) # random color generator
            point_color = '#%02X%02X%02X' % (r(),r(),r()) # random color generator
        viewer.add_points(points, size=point_size,face_color=point_color,visible=False,name=phenotype_layer)

    if overlay is not None:
        # categories under investigation
        if overlay_category is None:
            available_phenotypes = list(adata.obs[overlay].unique())
        else:
            available_phenotypes = overlay_category

        # Run the function on all phenotypes
        for i in available_phenotypes:
            add_phenotype_layer (adata=adata, overlay=overlay,
                                    phenotype_layer=i, x=x_coordinate, y=y_coordinate, viewer=viewer,
                                    point_size=point_size,point_color=point_color)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Wed Apr  7 17:46:29 2021
# @author: Ajit Johnson Nirmal
"""
!!! abstract "Short Description"
    `sm.pl.foldchange`: The Function allows users to visualize foldchange in abundance of celltypes between samples/ROI's. 
    Run `sm.tl.foldchange` first to compute the foldchange.

## Function
"""

# lib
import seaborn as sns; sns.set(color_codes=True)
import matplotlib
import matplotlib.pyplot as plt 
import numpy as np
from pandas.plotting import parallel_coordinates
sns.set_style("white")

# Function
def foldchange (adata, label='foldchange', 
                p_val=0.05, nonsig_color='grey',subset_xaxis=None,subset_yaxis=None,
                cmap = 'vlag', log=True,center=0, 
                method='heatmap', invert_axis=None,
                parallel_coordinates_color=None,matplotlib_bbox_to_anchor=(1.04,1),
                matplotlib_legend_loc='upper left',xticks_rotation=90,
                return_data = False,
                **kwargs):
    """
Parameters:
    adata : Anndata object

    label : strong, optional  
        label used when running `sm.tl.foldchange`.

    p_val : float, optional  
        p_val cut-off above which is considered not-significant. The cells containing
        non-significant changes will be highlighted in the heatmap.

    nonsig_color : string, optional  
        Color used to highlight non-significant fold changes in the heatmap.

    subset_xaxis : list, optional  
        Subset x-axis before plotting. Pass in a list of categories. eg- subset_xaxis = ['CelltypeA', 'CellTypeB']. 

    subset_yaxis : list, optional  
        Subset y-axis before plotting. Pass in a list of categories. eg- subset_yaxis = ['ROI_1', 'ROI_5']. 

    cmap : string, optional  
        Color map. Can be a name or a Colormap instance (e.g. 'magma', 'viridis').

    log : bool, optional  
        Convert foldchange to log2 scale.

    center : float, optional  
        The center value to be used in heatmap.

    method : string, optional  
        Two methods are available for plotting the foldchanges  
        a) Heatmap: Use `heatmap`  
        b) parallel coordinates plot : Use `parallel_coordinates`  

    invert_axis : bool, optional  
        Flip the axis of the plot.

    parallel_coordinates_color : list, optional  
        Custom colors for each category.

    matplotlib_bbox_to_anchor : tuple, optional  
        Bounding box argument used along with matplotlib_legend_loc to control
        the legend location when using the matplotlib method.

    matplotlib_legend_loc : TYPE, optional  
        Location of legend used along with matplotlib_bbox_to_anchor to control
        the legend location when using the matplotlib method.

    xticks_rotation : int, optional  
        Angle the x-axis ticks.

    return_data: bool, optional  
        Return the final data used for plotting.

    **kwargs : Additional keyword arguments passed to:  
        a) sns.clustermap  
        b) pandas.parallel_coordinates  

Returns:
    Plot:  
        Data used for the plot if `return_data = True`
    
Example:
```python
    # Heatmap of foldchnage  
    sm.pl.foldchange (adata, label='foldchange', method='heatmap',
                     p_val=0.05, nonsig_color='grey',
                     cmap = 'vlag', log=True, center=0, linecolor='black',linewidths=0.7,
                     vmin=-5, vmax=5, row_cluster=False)
    
    # Parallel_coordinates plot of the foldchanges
    foldchange (adata, label='foldchange', 
                log=True, method='parallel_coordinates', invert_axis=True,
                parallel_coordinates_color=['black','blue','green','red','#000000'],
                matplotlib_bbox_to_anchor=(1.04,1),
                matplotlib_legend_loc='upper left',
                xticks_rotation=90,
                return_data = False
```
    """
        
    # set color for heatmap
    #cmap_updated = copy.copy(matplotlib.cm.get_cmap(cmap))
    cmap_updated = matplotlib.cm.get_cmap(cmap)
    cmap_updated.set_bad(color=nonsig_color)
    
    
    # get the data
    fc = adata.uns[str(label)+'_fc']
    p = adata.uns[str(label)+'_pval']
        
    #fold
    fold = fc.copy()
    p_mask = p.copy()
    
    # reference image
    ref = fold.index.name
       
    # log
    if log is True:
        fold = np.log2(fold)
    
    # create a mask for non-sig values
    p_mask[p_mask > p_val] = np.nan
    
    # subset x axis data
    if subset_xaxis is not None:
        if isinstance (subset_xaxis, str):
            subset_xaxis = [subset_xaxis]
        fold = fold [subset_xaxis]
        p_mask = p_mask [subset_xaxis]
        #reorder
        
    # subset y axis data
    if subset_yaxis is not None:
        if isinstance (subset_yaxis, str):
            subset_yaxis = [subset_yaxis]
        fold = fold.loc [subset_yaxis]
        p_mask = p_mask.loc [subset_yaxis]
        #reorder
        
    # invert axis if user requests
    if invert_axis is True:
        fold = fold.T
        p_mask = p_mask.T
    
    #mask
    mask = p_mask.isnull() # identify the NAN's for masking 
    
    if method == 'heatmap':
        # heatmap of the foldchange
        #g= sns.clustermap(fold, cmap=cmap, mask=mask, center=center, col_cluster=False, row_cluster=False)
        g= sns.clustermap(fold, cmap=cmap, mask=mask, center=center, **kwargs)
        plt.suptitle('reference: '+ str(ref))
        plt.setp(g.ax_heatmap.get_xticklabels(), rotation=xticks_rotation)
        plt.tight_layout()
        
    
    if method == 'parallel_coordinates':
        fold['sample'] = fold.index
        # plotting
        fig, axes = plt.subplots()
        if parallel_coordinates_color is not None:
            parallel_coordinates(fold, 'sample', color=parallel_coordinates_color, **kwargs)
        else:
            #parallel_coordinates(fold, 'sample', colormap=cmap_updated)
            parallel_coordinates(fold, 'sample', colormap=cmap_updated, **kwargs)
        axes.grid(False)
        plt.legend(bbox_to_anchor=matplotlib_bbox_to_anchor, loc=matplotlib_legend_loc)
        plt.axhline(y=0, color='black', linestyle='-')
        plt.xticks(rotation = xticks_rotation)
        plt.suptitle('reference: '+ str(ref))
        fig.tight_layout()
        
    # return data
    if return_data is True:
        return fold
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Tue May 12 23:47:52 2020
# @author: Ajit Johnson Nirmal
"""
!!! abstract "Short Description"
    `sm.pl.gate_finder`: The function opens the OME-TIFF image inside Napari and overlays points to help with the 
    identifying manual gates for each marker. Use the `sm.pp.rescale` function to apply the identified gates to your data.

## Function
"""

try:
    import napari
except:
    pass

import pandas as pd
import tifffile as tiff
import numpy as np

import dask.array as da
from dask.cache import Cache
import zarr
import os

cache = Cache(2e9)  # Leverage two gigabytes of memory
cache.register()

def gate_finder (image_path, adata, marker_of_interest, from_gate = 6, to_gate = 8, increment = 0.1,
                 markers=None, channel_names = 'default', flip_y=True,
                 x_coordinate='X_centroid',y_coordinate='Y_centroid',
                 point_size=10,imageid='imageid',subset=None,seg_mask=None,**kwargs):
    """
Parameters:
    image_path : string  
        Location to the image file.

    adata : Ann Data Object  

    marker_of_interest : string  
        Marker for which gate is to be defined e.g. 'CD45'.

    from_gate : int, optional  
        Start value gate of interest.

    to_gate : int, optional  
        End value of the gate of interest.
        
    flip_y : bool, optional  
        Flip the Y-axis if needed. Some algorithms output the XY with the Y-coordinates flipped.
        If the image overlays do not align to the cells, try again by setting this to `False`.

    increment : float, optional  
        Increments between the start and end values.

    markers : string, optional  
        Additional markers to be included in the plot for evaluation.

    channel_names : list, optional  
        List of channels in the image in the exact order as image. The default is `adata.uns['all_markers']`

    x_coordinate : string, optional  
        X axis coordinate column name in AnnData object.

    y_coordinate : string, optional  
        Y axis coordinate column name in AnnData object.

    point_size : int, optional  
        point size in the napari plot.

    imageid : string, optional  
        Column name of the column containing the image id.

    subset : string, optional  
        imageid of a single image to be subsetted for analyis.

    seg_mask : string, optional  
        Location to the segmentation mask file.

    **kwargs  
        Other arguments that can be passed to napari viewer.

Example:
```python
    image_path = '/Users/aj/Desktop/ptcl_tma/image.ome.tif'
    sm.pl.gate_finder (image_path, adata, marker_of_interest='CD45',
                 from_gate = 6, to_gate = 8, increment = 0.1,
                 markers=['DNA10'], channel_names = 'default',
                 x_coordinate='X_position',y_coordinate='Y_position',point_size=10,
                 subset= '77', seg_mask=None)
```
    """

    # If no raw data is available make a copy
    if adata.raw is None:
        adata.raw = adata

    # Copy of the raw data if it exisits
    if adata.raw is not None:
        adata.X = adata.raw.X

    # Make a copy of the data with the marker of interest
    data = pd.DataFrame(np.log1p(adata.X), columns = adata.var.index, index= adata.obs.index)[[marker_of_interest]]

    # Generate a dataframe with various gates
    def gate (g, d):
        dd = d.values
        dd = np.where(dd < g, np.nan, dd)
        np.warnings.filterwarnings('ignore')
        dd = np.where(dd > g, 1, dd)
        dd = pd.DataFrame(dd, index = d.index, columns = ['gate-' + str(g)])
        return dd

    # Identify the list of increments
    inc = list(np.arange (from_gate, to_gate, increment))
    inc = [round(num,3) for num in inc]

    # Apply the function
    r_gate = lambda x: gate(g=x, d=data) # Create lamda function
    gated_data = list(map(r_gate, inc)) # Apply function
    # Concat all the results into a single dataframe
    gates = pd.concat(gated_data, axis=1)
    
    # Plot only the Image that is requested
    if subset is not None:
        adata = adata[adata.obs[imageid] == subset]
        
    # Recover the channel names from adata
    if channel_names == 'default':
        channel_names = adata.uns['all_markers']
    else:
        channel_names = channel_names

    # if markers is a string convert to list
    if isinstance(markers, str):
        markers = [markers]
        
    # Index of the marker of interest and corresponding names
    if markers is not None:
        markers.extend([marker_of_interest])
        idx = np.where(np.isin(channel_names,markers))[0]
        channel_names = [channel_names[i] for i in idx]
    else:
        idx = list(range(len(channel_names)))
        channel_names = channel_names


    # Load the segmentation mask
    if seg_mask is not None:
        seg_m = tiff.imread(seg_mask)
        if (len(seg_m.shape) > 2) and (seg_m.shape[0] > 1):
            seg_m = seg_m[0]
        

    ##########################################################################
    # Visulaisation using Napari
    
    # load OME TIFF
    if os.path.isfile(image_path) is True: 
        # Load the image    
        image = tiff.TiffFile(image_path, is_ome=False)
        z = zarr.open(image.aszarr(), mode='r') # convert image to Zarr array
        # Identify the number of pyramids and number of channels
        n_levels = len(image.series[0].levels) # pyramid  
        # If and if not pyramids are available
        if n_levels > 1:
            pyramid = [da.from_zarr(z[i]) for i in range(n_levels)]
            multiscale = True
        else:
            pyramid = da.from_zarr(z)
            multiscale = False   
        # subset channels of interest
        if markers is not None:
            if n_levels > 1:
                for i in range(n_levels-1):
                    pyramid[i] = pyramid[i][idx, :, :]
                n_channels = pyramid[0].shape[0] # identify the number of channels
            else:
                pyramid = pyramid[idx, :, :]
                n_channels = pyramid.shape[0] # identify the number of channels
        else:
            if n_levels > 1:
                n_channels = pyramid[0].shape[0]
            else:
                n_channels = pyramid.shape[0]\
        # check if channel names have been passed to all channels
        if channel_names is not None:
            assert n_channels == len(channel_names), (
                f'number of channel names ({len(channel_names)}) must '
                f'match number of channels ({n_channels})'
            )

        # Load the viewer
        viewer = napari.view_image(
        pyramid,
        channel_axis = 0,
        multiscale=multiscale,
        name = None if channel_names is None else channel_names,
        visible = False, **kwargs)


    # Operations on the ZARR image
    # check the format of image
    if os.path.isfile(image_path) is False: 
        #print(image_path)
        viewer = napari.Viewer()
        viewer.open(image_path, multiscale=True,
                    visible=False,
                    name = None if channel_names is None else channel_names)

    # Add the seg mask
    if seg_mask is not None:
        viewer.add_labels(seg_m, name='segmentation mask', visible=False)

    # subset the gates to include only the image of interest
    gates = gates.loc[adata.obs.index,]

    # Add gating layer
    def add_phenotype_layer (adata, gates, phenotype_layer,x,y,viewer,point_size):
        cells = gates[gates[phenotype_layer] == 1].index
        coordinates = adata[cells]
        # Flip Y axis if needed
        if flip_y is True:
            coordinates = pd.DataFrame({'y': coordinates.obs[y],'x': coordinates.obs[x]})
        else:  
            coordinates = pd.DataFrame({'x': coordinates.obs[x],'y': coordinates.obs[y]})
        #points = coordinates.values.tolist()
        points = coordinates.values
        #import time
        #start = time.time()
        viewer.add_points(points, size=point_size, face_color='white',visible=False,name=phenotype_layer)
        #stop = time.time()
        #print(stop-start)
        

    # Run the function on all gating layer
    for i in gates.columns:
        add_phenotype_layer (adata=adata, gates=gates, 
                             phenotype_layer=i, x=x_coordinate, y=y_coordinate, 
                             viewer=viewer, point_size=point_size)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Tue Apr  6 22:05:14 2021
# @author: Ajit Johnson Nirmal
"""
!!! abstract "Short Description"
    `sm.pl.pie`: The function allows users to plot a pie plot for any categorical column of interest. 

## Function
"""

# Lib
import pandas as pd
import matplotlib.pyplot as plt
import numpy as np

# Function
def pie (adata, phenotype='phenotype', group_by='imageid', ncols=None,
         subset_phenotype=None, subset_groupby=None,
         label='auto', title='auto', colors=None, autopct='%1.1f%%',
         legend=False,legend_loc='upper right',
         wedgeprops = {'linewidth': 0}, return_data=False, **kwargs):
    """
Parameters:

    adata : AnnData object

    phenotype : string, optional  
        Column contaning the cell-type inforamtion or any categorical data to be displayed
        in the form of a pie plot.

    group_by : string, optional  
        Column that contains inforamtion on data groupings which leads to generation of
        pie plot for each group (e.g. image-id). If `None` is passed,
        the entire data is considered as a single group.

    ncols : int, optional  
        In case group_by is used, a grid of plots are returned. This paramenter
        controls the number of columns in that grid.

    subset_phenotype : list, optional  
        User can subset a list of categories within `phenotype` before plotting.

    subset_groupby : list, optional  
        User can subset a list of categories within `group_by` before plotting.

    label : list, optional  
        A list of strings providing the labels for each wedge.

    title : string, optional  
        If `None`, the title of the pieplot is not plotted.

    colors : list, optional  
        A sequence of colors through which the pie chart will cycle. If None, will use the 
        colors in the currently active cycle.

    legend : bool, optional  
        If True, color legends are plotted seperately.

    legend_loc: string, optional  
        Place a legend on the Axes.

    autopct : None or str or callable, optional  
        If not None, is a string or function used to label the wedges with their numeric value. 
        The label will be placed inside the wedge. If it is a format string, 
        the label will be fmt % pct. If it is a function, it will be called.

    wedgeprops : dict, optional  
        Dict of arguments passed to the wedge objects making the pie. For example, you can pass in 
        wedgeprops = {'linewidth': 3} to set the width of the wedge border lines equal to 3. 
        For more details, look at the doc/arguments of the wedge object. By default clip_on=False.

    return_data : bool, optional  
        Returns the data used for plotting.

    **kwargs :  
        Keyword arguments to pass on to `matplotlib.pyplot.pie`.

Returns:  
    Returns data used for plotting if `return_data = True`
    
Example:
```python
    # pie plot showing stromal tumor content among the different samples
    sm.pl.pie (adata, phenotype='Tumor_Stroma', group_by='imageid', 
               autopct='%1.1f%%',
               textprops={'fontsize': 8, 'color': '#1d3557', 'fontweight': 'bold'},
               ncols=5, label=None, title=None, 
               colors=['#a8dadc','#e63946'], 
               wedgeprops = {'linewidth': 0.8})
```
    """
    
    
    # convert subset into list
    if subset_phenotype is not None:
        if isinstance (subset_phenotype, str): 
            subset_phenotype = [subset_phenotype]
    if subset_groupby is not None:
        if isinstance (subset_groupby, str):
            subset_groupby = [subset_groupby]
    
    # create copy of the required data
    if group_by is not None:
        data = adata.obs[[phenotype,group_by]]
    else:
        data = adata.obs[[phenotype]]

    # subset data if needed
    if subset_groupby is not None:
        data = data[data[group_by].isin(subset_groupby)]
        data[group_by] = data[group_by].astype('str').astype('category')
        data[group_by] = data[group_by].cat.reorder_categories(subset_groupby)
        data = data.sort_values(group_by)
    if subset_phenotype is not None:
        data = data[data[phenotype].isin(subset_phenotype)]
        data[phenotype] = data[phenotype].astype('str').astype('category')      
        data[phenotype] = data[phenotype].cat.reorder_categories(subset_phenotype)
        data = data.sort_values(phenotype)
    if group_by and phenotype is not None:
        data = data.sort_values([phenotype, group_by])
        
        
    # calculate the proportion
    if group_by is None:
        prop = data[phenotype].value_counts().reset_index(inplace=False)
        prop.columns = [phenotype, 'value']
        prop['group_by'] = phenotype
        labels = np.unique(prop[phenotype])

    else:
        # if group_by is provided
        prop = pd.DataFrame(data.groupby([group_by,phenotype]).size()).reset_index(inplace=False)
        prop.columns = ['group_by',phenotype,'value']
        labels = np.unique(prop[phenotype])
        #
        if ncols is not None:
            g = prop.groupby('group_by')
            rows = int(np.ceil(len(g)/ncols))
        else:
            g = prop.groupby('group_by')
            rows = 1
            ncols = len(g)
    
    # remove label if requested 
    if label == 'auto':
        label = labels
    elif label is None:
        label = None
    else:
        label = label
    
    
    # plot
    if group_by is None:
        fig, ax = plt.subplots()
        ax.pie(prop.value, labels=label,colors=colors, wedgeprops = wedgeprops)
        #ax.pie(prop.value, labels=label,colors=colors, wedgeprops = wedgeprops, **kwargs)
        if title is None:
            pass
        else:
            ax.set_title(phenotype)
    else:
        # plot the figure
        # Ground work for removing unwanted axes
        total_axes = list(range(ncols * rows))
        required_axes = list(range(len(np.unique(prop['group_by']))))
        final_axes = list(set(total_axes) ^ set(required_axes))      
        # Plot
        fig, axes = plt.subplots(ncols=ncols, nrows=rows)
        for (c, grp), ax in zip(g, axes.flat):
            ax.pie(grp.value, labels=label, colors=colors, wedgeprops =wedgeprops)
            #ax.pie(grp.value, labels=label, colors=colors, wedgeprops = wedgeprops, **kwargs)
            if title is None:
                pass
            else:
                ax.set_title(c)        
        # removing unwanted axis
        for i in final_axes:
            fig.delaxes(axes.flatten()[i])
        
        if legend is True:
            plt.legend(labels, loc=legend_loc, framealpha=1)
            
    plt.show()
    

    # return data
    if return_data is True:
        return prop
    
    


# import scimap as sm
from .. import tools as tl
from .. import preprocessing as pp
from .. import plotting as pl
from .. import helpers as hl

import sys
import argparse
import pathlib
import textwrap
from joblib import Parallel, delayed

def mcmicro_wrap(argv=sys.argv):
    
    parser = argparse.ArgumentParser()

    parser.add_argument(
        'csv',
        help='single-cell quantification table (CSV file) generated by mcmicro',
    )
    parser.add_argument(
        '-o', '--output', default='.',
        help='output directory'
    )

    args = parser.parse_args(argv[1:])

    run_result = Parallel(n_jobs=-1, verbose=1)(
        delayed(clustering)([None, args.csv, '-o', args.output, '--clustering-method', m]) 
        for m in ['spatial', 'kmeans', 'phenograph', 'leiden']
    )

    print(run_result)
    
    merge([None, args.output, '-o', args.output, '-d', '--csv'])

    # move pdf files to output / plots
    output_dir = pathlib.Path(args.output)
    pdfs = sorted(output_dir.rglob('*.pdf'))
    plots_dir = output_dir / 'plots'
    plots_dir.mkdir(exist_ok=True)

    if len(pdfs) > 0:
        print(textwrap.indent(
            'Moving pdf plots:\n',
            '    '
        ))
    for p in pdfs:
        if plots_dir not in p.parents:
            move_to = plots_dir / p.parent.name / p.name
            print(textwrap.indent(
                f'{str(p)}\n    -> {str(move_to)}',
                '    '
            ))
            move_to.parent.mkdir(exist_ok=True, parents=True)
            p.replace(move_to)
    print()
    return 0

def clustering(argv=sys.argv):

    parser = argparse.ArgumentParser()

    parser.add_argument(
        'csv',
        help='single-cell quantification table (CSV file) generated by mcmicro',
    )
    parser.add_argument(
        '-o', '--output', default='.',
        help='output directory'
    )
    clustering_methods = ['all', 'spatial', 'kmeans', 'phenograph', 'leiden', 'pass']
    parser.add_argument(
        '--clustering-method', default=['all'], choices=clustering_methods, nargs='+', 
        help='choice of clustering algorithms, "pass": do not run any clustering methods; "all": run all clustering methods; default: "all"'
    )

    args = parser.parse_args(argv[1:])

    _output_dir = pathlib.Path(args.output)
    mcmicro_csv_path = args.csv
    methods = set(args.clustering_method)

    assert '.csv' in mcmicro_csv_path, 'input file must be a csv file'
    
    if 'all' in methods:
        methods = clustering_methods[1:-1]

    if 'pass' in methods:
        pp.mcmicro_to_scimap(
            feature_table_path = mcmicro_csv_path, 
            output_dir = str(_output_dir)
        )
        return

    for method in methods:

        output_dir = _output_dir / method
        
        pp.mcmicro_to_scimap(
            feature_table_path = mcmicro_csv_path, 
            output_dir = str(output_dir)
        )

        adata_path = pathlib.Path(output_dir) / f'{pathlib.Path(mcmicro_csv_path).stem}.h5ad'

        if method == 'spatial':
            # Spatial clustering
            tl.spatial_expression(
                adata = str(adata_path), 
                output_dir = output_dir
            )
            tl.spatial_cluster(
                adata = str(adata_path), 
                df_name = "spatial_expression",
                output_dir = output_dir
            )
        
        else:
            # Expression clustering
            tl.cluster(
                adata = str(adata_path), 
                method = method,
                output_dir = output_dir
            )

            # Expression clustering plots
            pl.cluster_plots(
                adata = str(adata_path), 
                group_by  = method,
                output_dir = output_dir
            )
    return 0

def merge(argv=sys.argv):
    
    parser = argparse.ArgumentParser()

    parser.add_argument(
        'directory',
        help='recursively search for .h5ad files to merge',
    )
    parser.add_argument(
        '-o', '--output', default='.',
        help='output directory'
    )
    parser.add_argument(
        '-d', '--delete-merged', default=False, action='store_true',
        help='delete found input files after merging; default: False'
    )
    parser.add_argument(
        '--csv', default=False, action='store_true',
        help='output csv version of the merged h5ad file; default: False'
    )

    args = parser.parse_args(argv[1:])

    input_dir = pathlib.Path(args.directory)
    output_dir = pathlib.Path(args.output)
    delete_after = args.delete_merged
    output_csv = args.csv

    input_files = sorted(input_dir.rglob('*.h5ad'))
    output_file = output_dir / input_files[0].name
    
    if len(input_files) == 0:
        print(f'0 .h5ad found in {str(input_dir)}')
        return 1

    print_text = '''
        Merging:

        {}

        Writing:

        {}'''
    print_text = textwrap.dedent(print_text).format(
        "\n".join([str(p) for p in input_files]),
        str(output_file)
    )
    print(textwrap.indent(print_text, '    '))
    # Merge data
    hl.merge_adata_obs(
        adata = [
            str(p)
            for p in input_files
        ],
        output_dir = output_dir
    )
    if output_csv == True:
        hl.scimap_to_csv(
            adata = str(output_file),
            output_dir = output_dir
        )
        print(textwrap.indent(
            str(output_file.parent / f'{output_file.stem}.csv\n'),
            '    '
        ))
    

    if delete_after == True:
        if output_file in input_files:
            idx = input_files.index(output_file)
            input_files.pop(idx)
        print(
            textwrap.indent(
                'Deleting:\n\n' + "\n".join([str(f) for f in input_files]),
                '    '
            )
        )
        for f in input_files:
            f.unlink()
        print()
    return 0


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Sat May 28 19:13:24 2022
# @author: Ajit Johnson Nirmal
# UMAP Function

"""
!!! abstract "Short Description"
    `sm.tl.umap`: The function allows users to perform dimensionality reduction using UMAP 
## Function
"""

# libs
import umap as um
import numpy as np


# function
def umap (adata, use_layer=None, use_raw=False, log=False,
          n_neighbors=15, n_components=2, metric='euclidean',min_dist=0.1, random_state=0, 
          label='umap', **kwargs):
    """
Parameters:

    adata : AnnData Object  
    
    use_layer : string, optional  
        Pass name of any `Layer` in AnnData. The default is `None` and `adata.X` is used.
        
    use_raw : bool, optional  
        If set to `True`, values in `adata.raw.X` will be used. The default is False.
        
    log : bool, optional  
        If set to `True`, the data will natural log transformed using `np.log1p()`. The default is False.
        
    n_neighbors : int, optional  
        The size of local neighborhood (in terms of number of neighboring sample points) 
        used for manifold approximation. Larger values result in more global views of the manifold, 
        while smaller values result in more local data being preserved. In general 
        values should be in the range 2 to 100. The default is 15.
        
    n_components : int, optional  
        The dimension of the space to embed into. This defaults to 2 to provide easy visualization, 
        but can reasonably be set to any integer value in the range 2 to 100. The default is 2.
        
    metric : TYPE, optional  
        The metric to use to compute distances in high dimensional space. 
        Check `https://umap-learn.readthedocs.io/en/latest/api.html` for all available 
        options. The default is 'euclidean'.
        
    min_dist : float, optional  
        The effective minimum distance between embedded points. Smaller values will 
        result in a more clustered/clumped embedding where nearby points on the manifold 
        are drawn closer together, while larger values will result on a more even 
        dispersal of points. The value should be set relative to the spread value, 
        which determines the scale at which embedded points will be spread out. The default is 0.1.
        
    random_state : int, optional  
        If int, random_state is the seed used by the random number generator; 
        If RandomState instance, random_state is the random number generator; 
        If None, the random number generator is the RandomState instance used by np.random. The default is 0.
        
    label : string, optional  
        Key used to save the embeddings. 
        Check `adata.obsm[label]` for results. The default is 'umap'.
    
    **kwargs : Other `umap` parameters. Check `https://umap-learn.readthedocs.io/en/latest/api.html`  

Returns:

    adata : Modified AnnData object
        Embedding stored in `adata.obsm[label]`.
    
Example:
    ```python
# Run UMAP
adata = sm.tl.umap(adata)
    
# plot results
sm.pl.umap(adata)
    ```

    """

    # adata_layer=None;use_raw=False;log=False;n_neighbors=15;n_components=2;metric='euclidean';min_dist=0.1;
    # random_state=0;
    # load data
    if use_layer is not None:
        data = adata.layers[use_layer]
    elif use_raw is True:
        data = adata.raw.X
    else:
        data = adata.X
    
    # log the data if user requests
    if log is True:
        data = np.log1p(data)
    
        
    # embedding
    embedding = um.UMAP(n_neighbors=n_neighbors,
                          n_components=n_components,
                          metric=metric,
                          min_dist=min_dist,
                          random_state=random_state).fit_transform(data)
    
    # plot
    #plt.scatter(embedding[:, 0], embedding[:, 1], s=5)
    
    # return data
    adata.obsm[label] = embedding
    return adata



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
!!! abstract "Short Description"
    `sm.pl.merge_adata_obs`: The function allows users to combine multiple anndata objects that are from the same image or dataset 
    but differ in the stored metadata `.obs`. For example, multiple clustering methods could have been run in parallel
    leading to generation of multiple anndata objects. This allows to merge them togeather into a single object. Please
    note this cannot be used to merge anndata objects from different images/ datasets. 

## Function
"""

import argparse
import sys
import pathlib
import pandas as pd
import anndata as ad


def main(argv=sys.argv):
    parser = argparse.ArgumentParser(
        description='The function allows users to combine multiple anndata objects that are from the same image or dataset but differ in the stored metadata'
    )
    parser.add_argument(
        '--adata', required=True, nargs='*',
        help='AnnData object loaded into memory or path to AnnData object.'
    )
    parser.add_argument(
        '--output_dir', type=str, required=False, default=None,
        help='Path to output directory.'
    )
    args = parser.parse_args(argv[1:])
    print(vars(args))
    merge_adata_obs(**vars(args))
    
    

def merge_adata_obs (adata, output_dir=None):
    
    
    """
Parameters:
    adata : List of AnnData object loaded into memory or list of path to AnnData object.
        
    output_dir : string, optional  
        Path to output directory.

Returns:
    AnnData :   
        Combined anndata object.
        
Example:
```python
    # combining multiple anndata objects in memory
    adata = sm.hl.merge_adata_obs (adata = [bdata, cdata])
    
    # combining multiple anndata objects by providing their saved loaction
    adata = sm.hl.merge_adata_obs (adata = ['path to bdata.h5ad', 'path to cdata.h5ad'])
```
    """
    
    #adata = ["/Users/aj/Downloads/mcmicro_output.h5ad", "/Users/aj/Downloads/mcmicro_output_1.h5ad"]
    
    
    # Convert to list of anndata objects
    if isinstance (adata, list):
        adata = adata
    else:
        adata = [adata]
    
    # Resoluve the OBS and UNS section
    if isinstance(adata[0], str):
        df = pd.DataFrame()
        uns_count = []
        for i in adata:
            tmp = ad.read(i)
            # OBS
            tmp_df = tmp.obs
            df = pd.concat([df, tmp_df], axis=1)
            # UNS
            uns_count.append(len(tmp.uns))
        # Keep non-duplicate obs columns
        df = df.loc[:,~df.columns.duplicated()]
        # Highest UNS
        uns_index = [i for i, j in enumerate(uns_count) if j == max(uns_count)][0]
    else:
        df = pd.DataFrame()
        uns_count = []
        for i in adata:
            # OBS
            tmp_df = i.obs
            df = pd.concat([df, tmp_df], axis=1)
            # UNS
            uns_count.append(len(i.uns))
        # Keep non-duplicate obs columns
        df = df.loc[:,~df.columns.duplicated()]
        # Highest UNS
        uns_index = [i for i, j in enumerate(uns_count) if j == max(uns_count)][0]
    
    
    # create the final anndata object
    # Load the data 
    if isinstance(adata[0], str):
        final_adata = ad.read(adata[uns_index])
    else:
        final_adata = adata[uns_index]
        
    # reindex
    df = df.reindex (final_adata.obs.index)
    
    # replace obs
    final_adata.obs = df
    
    # Find name of file
    image_path = pathlib.Path(adata[0])    
    
    # Save data if requested
    if output_dir is not None:
        output_dir = pathlib.Path(output_dir)
        output_dir.mkdir(exist_ok=True, parents=True)
        final_adata.write(output_dir / image_path.name)
    else:    
        # Return data
        return final_adata
    
    

if __name__ == '__main__':
    main()  
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Wed Aug 19 15:00:39 2020
# @author: Ajit Johnson Nirmal
"""
!!! abstract "Short Description"
    `sm.tl.spatial_aggregate`: The function allows users to find regions of aggregration of similar cells.
    Use the `purity` parameter to fine-tune percent of similar cells within a given `radius`. 

## Function
"""

# Import library
import pandas as pd
import numpy as np
from sklearn.neighbors import BallTree

# Function
def spatial_aggregate (adata, x_coordinate='X_centroid',y_coordinate='Y_centroid',
                       purity = 60, phenotype='phenotype', method='radius', radius=30, knn=10, 
                       imageid='imageid',subset=None,label='spatial_aggregate'):
    """
    
Parameters:
    adata : AnnData object

    x_coordinate : float, required  
        Column name containing the x-coordinates values.

    y_coordinate : float, required  
        Column name containing the y-coordinates values.

    purity : int, optional  
        Supply a value between 1 to 100. It is the percent purity of neighbouring cells.
        For e.g. if 60 is chosen, every neighbourhood is tested such that if a 
        particular phenotype makes up greater than 60% of the total 
        population it is annotated to be an aggregate of that particular phenotype.

    phenotype : string, required  
        Column name of the column containing the phenotype information. 
        It could also be any categorical assignment given to single cells.

    method : string, optional  
        Two options are available: a) 'radius', b) 'knn'.
        a) radius - Identifies the neighbours within a given radius for every cell.
        b) knn - Identifies the K nearest neigbours for every cell.

    radius : int, optional  
        The radius used to define a local neighbhourhood.

    knn : int, optional  
        Number of cells considered for defining the local neighbhourhood.

    imageid : string, optional  
        Column name of the column containing the image id.

    subset : string, optional  
        imageid of a single image to be subsetted for analyis.

    label : string, optional  
        Key for the returned data, stored in `adata.obs`.

Returns:
    adata : AnnData object
        Updated AnnData object with the results stored in `adata.obs['spatial_aggregate']`.
        
        
Example:
```python
    # Running the radius method
    adata = sm.tl.spatial_aggregate (adata, x_coordinate='X_centroid',y_coordinate='Y_centroid',
                        phenotype='phenotype', method='radius', radius=30, purity = 60,
                        imageid='imageid',subset=None,label='spatial_aggregate_radius')
    
    # Running the knn method
    adata =  sm.tl.spatial_aggregate (adata, x_coordinate='X_centroid',y_coordinate='Y_centroid',
                        phenotype='phenotype', method='knn', knn=10, purity = 60,
                        imageid='imageid',subset=None,label='spatial_aggregate_knn')
```
    """
    
    # Error statements
    #if purity < 51:
    #    raise ValueError('purity should be set to a value greater than 50')
        
    def spatial_aggregate_internal (adata_subset, x_coordinate,y_coordinate,phenotype,purity,
                                    method,radius,knn,imageid,subset,label):
    

        # Create a DataFrame with the necessary inforamtion
        data = pd.DataFrame({'x': adata_subset.obs[x_coordinate], 'y': adata_subset.obs[y_coordinate], 'phenotype': adata_subset.obs[phenotype]})
        
        # Identify neighbourhoods based on the method used
        # a) KNN method
        if method == 'knn':
            print("Identifying the " + str(knn) + " nearest neighbours for every cell")
            tree = BallTree(data[['x','y']], leaf_size= 2)
            ind = tree.query(data[['x','y']], k=knn, return_distance= False)
            neighbours = pd.DataFrame(ind.tolist(), index = data.index) # neighbour DF
            neighbours.drop(0, axis=1, inplace=True) # Remove self neighbour
        
        # b) Local radius method
        if method == 'radius':
            print("Identifying neighbours within " + str(radius) + " pixels of every cell")
            kdt = BallTree(data[['x','y']], leaf_size= 2) 
            ind = kdt.query_radius(data[['x','y']], r=radius, return_distance=False)
            for i in range(0, len(ind)): ind[i] = np.delete(ind[i], np.argwhere(ind[i] == i))#remove self
            neighbours = pd.DataFrame(ind.tolist(), index = data.index) # neighbour DF
            
        # Map phenotype
        phenomap = dict(zip(list(range(len(ind))), data['phenotype'])) # Used for mapping
        
        # Loop through (all functionized methods were very slow)
        for i in neighbours.columns:
            neighbours[i] = neighbours[i].dropna().map(phenomap, na_action='ignore')
        
        # Drop NA
        #n_dropped = neighbours.dropna(how='all')
        
        
        # Collapse all the neighbours into a single column
        n = pd.DataFrame(neighbours.stack(), columns = ["neighbour_phenotype"])
        n.index = n.index.get_level_values(0) # Drop the multi index
        n = pd.DataFrame(n)
        n['order'] = list(range(len(n)))
        
        # Merge with real phenotype
        n_m = n.merge(data['phenotype'], how='inner', left_index=True, right_index=True)
        n_m['neighbourhood'] = n_m.index
        n = n_m.sort_values(by=['order'])
        
        # Count the neighbours
        k = n.groupby(['neighbourhood','neighbour_phenotype']).size().unstack().fillna(0)
        k = k.div(k.sum(axis=1), axis=0)
        
        # Iteratte over all rows and find the column which passes the purity test
        #def col_name_mapper (row_data, purity):
        #    p = row_data[row_data >= purity/100]
        #    #phenotype_name = 'non-significant' if len(p.index) == 0 else p.index[0]
        #    phenotype_name = 'non-significant' if len(p.index) == 0 else p.idxmax()
        #    return phenotype_name
        # Apply the iteration function
        #aggregate_pheno = pd.DataFrame(k.apply(lambda x: col_name_mapper(row_data=x,purity=purity), axis=1))
        aggregate_pheno = pd.DataFrame(k[k>=purity/100].idxmax(axis=1).fillna('non-significant'))
        aggregate_pheno.columns = ['spatial_aggregate']
        
        # Return 
        return aggregate_pheno
    

    # Subset a particular image if needed
    if subset is not None:
        adata_list = [adata[adata.obs[imageid] == subset]]
    else:
        adata_list = [adata[adata.obs[imageid] == i] for i in adata.obs[imageid].unique()]
    
    # Apply function to all images and create a master dataframe
    # Create lamda function 
    r_spatial_aggregate_internal = lambda x: spatial_aggregate_internal(adata_subset=x,
                                                                          x_coordinate=x_coordinate,
                                                                          y_coordinate=y_coordinate,
                                                                          phenotype=phenotype,
                                                                          method=method,
                                                                          radius=radius,knn=knn,
                                                                          imageid=imageid,subset=subset,
                                                                          purity=purity,
                                                                          label=label) 
    all_data = list(map(r_spatial_aggregate_internal, adata_list)) # Apply function 
    
    # Merge all the results into a single dataframe    
    result = []
    for i in range(len(all_data)):
        result.append(all_data[i])
    result = pd.concat(result, join='outer')  
    
    # Reindex the cells
    result = result.fillna(0)
    result = result.reindex(adata.obs.index)
    
    # Add to adata
    adata.obs[label] = result
    
    # Return        
    return adata




# -*- coding: utf-8 -*-
# Created on Mon Mar  2 09:12:35 2020
# @author: Ajit Johnson Nirmal

"""
!!! abstract "Short Description"
    `sm.pp.mcmicro_to_scimap`: The function allows users to directly import the output from [mcmicro](https://mcmicro.org/) 
    into `scimap`.

## Function
"""

# Import library
import numpy as np
import anndata as ad
import pandas as pd
import argparse
import sys
import pathlib


def main(argv=sys.argv):
    parser = argparse.ArgumentParser(
        description='The function allows users to directly import the output from mcmicro'
    )
    parser.add_argument(
        '--feature_table_path', nargs='*', required=True, 
        help='List of path to the single-cell feature tables. Each Image should have a unique path supplied.'
    )
    parser.add_argument(
        '--remove_dna', action='store_true', required=False,  default=True,
        help='Remove the DNA channels from the final output. Looks for channels with the string dna in it.'
    )
    parser.add_argument(
        '--remove_string_from_name', type=str, required=False, default=None,
        help='Used to celan up channel names. If a string is given, that particular string will be removed from all marker names. If multiple images are passed, just use the string that appears in the first image.'
    )
    parser.add_argument(
        '--log', required=False, default=True,
        help='Log the data (log1p transformation will be applied).'
    )
    parser.add_argument(
        '--drop_markers', nargs='*', required=False, default=None,
        help='List of markers to drop from the analysis. e.g. ["CD3D", "CD20"]'
    )
    parser.add_argument(
        '--random_sample', type=int, required=False, default=None,
        help='Randomly sub-sample the data with the desired number of cells.'
    )
    parser.add_argument(
        '--unique_CellId', required=False, default=True,
        help='By default, the function creates a unique name for each cell/row by combining the `CellId` and `imageid`. If you wish not to perform this operation please pass `False`. The function will use whatever is under `CellId`. In which case, please be careful to pass unique `CellId` especially when loading multiple datasets togeather.'
    )
    parser.add_argument(
        '--CellId', type=str, required=False, default='CellID',
        help='Name of the column that contains the cell ID.'
    )
    parser.add_argument(
        '--split', type=str, required=False, default='X_centroid',
        help='To split the CSV into counts table and meta data, pass in the name of the column that immediately follows the marker quantification.'
    )
    parser.add_argument( 
        '--custom_imageid', type=str, required=False, default=None,
        help='Pass a user defined Image ID. By default the name of the CSV file is used.'
    )
    parser.add_argument(
        '--min_cells', type=int, required=False, default=None,
        help='If these many cells are not in the image, the image will be dropped. Particulary useful when importing multiple images.'
    )
    parser.add_argument(
        '--output_dir', type=str, required=False, default=None,
        help='Path to output directory.'
    )
    args = parser.parse_args(argv[1:])
    print(vars(args))
    mcmicro_to_scimap(**vars(args))




def mcmicro_to_scimap (feature_table_path,remove_dna=True,remove_string_from_name=None,
                        log=True,drop_markers=None,random_sample=None, unique_CellId=True,
                        CellId='CellID',split='X_centroid',custom_imageid=None,
                        min_cells=None, output_dir=None):
    """
Parameters:

    feature_table_path : list  
        List of path to the single-cell spatial feature tables. Each Image should have a unique path supplied.

    remove_dna : bool, optional  
        Remove the DNA channels from the final output. Looks for channels with the string 'dna' in it.

    remove_string_from_name : string, optional  
        Used to celan up channel names. If a string is given, that particular string will be removed from all marker names.
        If multiple images are passed, just use the string that appears in the first image.

    log : bool, optional  
        Log the data (log1p transformation will be applied).

    drop_markers : list, optional  
        List of markers to drop from the analysis. e.g. ["CD3D", "CD20"].

    random_sample : int, optional  
        Randomly sub-sample the data with the desired number of cells.

    CellId : string, optional  
        Name of the column that contains the cell ID.

    unique_CellId: bool, optional  
        By default, the function creates a unique name for each cell/row by combining the 
        `CellId` and `imageid`. If you wish not to perform this operation please pass `False`.
        The function will use whatever is under `CellId`. In which case, please be careful to pass unique `CellId`
        especially when loading multiple datasets togeather.  

    split : string, optional  
        To split the CSV into counts table and meta data, pass in the name of the column
        that immediately follows the marker quantification.

    custom_imageid: string, optional  
        Pass a user defined Image ID. By default the name of the CSV file is used.

    min_cells: int, optional  
        If these many cells are not in the image, the image will be dropped.
        Particulary useful when importing multiple images.

    output_dir: string, optional  
        Path to output directory. 

Returns:

    AnnData Object


Example:
```python
feature_table_path = ['/Users/aj/whole_sections/PTCL1_450.csv',
                  '/Users/aj/whole_sections/PTCL2_552.csv']
adata = sm.pp.mcmicro_to_scimap (feature_table_path, drop_markers= ['CD21', 'ACTIN'], random_sample=5000)
```
    """
    
    # feature_table_path list or string
    if isinstance(feature_table_path, str):
        feature_table_path = [feature_table_path]
    feature_table_path = [pathlib.Path(p) for p in feature_table_path]

    # Import data based on the location provided
    def load_process_data (image):
        # Print the data that is being processed
        print(f"Loading {image.name}")
        d = pd.read_csv(image)
        # If the data does not have a unique image ID column, add one.
        if 'imageid' not in d.columns:
            if custom_imageid is not None:
                imid = custom_imageid
            else:
                #imid = random.randint(1000000,9999999)
                imid = image.stem
            d['imageid'] = imid
        # Unique name for the data
        if unique_CellId is True:
            d.index = d['imageid'].astype(str)+'_'+d[CellId].astype(str)
        else:
            d.index = d[CellId]
            
        # move image id and cellID column to end
        cellid_col = [col for col in d.columns if col != CellId] + [CellId]; d = d[cellid_col]
        imageid_col = [col for col in d.columns if col != 'imageid'] + ['imageid']; d = d[imageid_col]
        # If there is INF replace with zero
        d = d.replace([np.inf, -np.inf], 0)
        # Return data
        return d
    # Apply function to all images and create a master dataframe
    r_load_process_data = lambda x: load_process_data(image=x) # Create lamda function
    all_data = list(map(r_load_process_data, list(feature_table_path))) # Apply function

    # Merge all the data into a single large dataframe
    for i in range(len(all_data)):
        all_data[i].columns = all_data[0].columns
    entire_data = pd.concat(all_data, axis=0, sort=False)
    
    # Randomly sample the data
    if random_sample is not None:
        entire_data = entire_data.sample(n=random_sample,replace=False)

    #Remove the images that contain less than a defined threshold of cells (min_cells)
    if min_cells is not None:
        to_drop = entire_data['imageid'].value_counts()[entire_data['imageid'].value_counts() < min_cells].index
        entire_data = entire_data[~entire_data['imageid'].isin(to_drop)]
        print('Removed Images that contained less than '+str(min_cells)+' cells: '+ str(to_drop.values))

    # Split the data into expression data and meta data
    # Step-1 (Find the index of the column with name Area)
    split_idx = entire_data.columns.get_loc(split)
    meta = entire_data.iloc [:,split_idx:]
    # Step-2 (select only the expression values)
    entire_data = entire_data.iloc [:,:split_idx]

    # Rename the columns of the data
    if remove_string_from_name is not None:
        entire_data.columns = entire_data.columns.str.replace(remove_string_from_name, '')

    # Save a copy of the column names in the uns space of ANNDATA
    markers = list(entire_data.columns)

    # Remove DNA channels
    if remove_dna is True:
        entire_data = entire_data.loc[:,~entire_data.columns.str.contains('dna', case=False)]

    # Drop unnecessary markers
    if drop_markers is not None:
        if isinstance(drop_markers, str):
            drop_markers = [drop_markers]
        entire_data = entire_data.drop(columns=drop_markers)

    # Create an anndata object
    adata = ad.AnnData(entire_data)
    adata.obs = meta
    adata.uns['all_markers'] = markers

    # Add log data
    if log is True:
        adata.raw = adata
        adata.X = np.log1p(adata.X)

    # Save data if requested
    if output_dir is not None:
        output_dir = pathlib.Path(output_dir)
        output_dir.mkdir(exist_ok=True, parents=True)
        imid = feature_table_path[0].stem
        adata.write(output_dir / f'{imid}.h5ad')
        #adata.write(str(output_dir) + '/' + imid + '.h5ad')
    else:    
        # Return data
        return adata


if __name__ == '__main__':
    main()


# -*- coding: utf-8 -*-
#Created on Mon Apr  4 11:12:59 2022
#@author: Ajit Johnson Nirmal
#Adding ROI to data via Napari.

"""
!!! abstract "Short Description"
    `sm.pl.addROI_image`: The function allows users to add ROIs to the data. 
    The function opens the image in a `napari` viewer and allows users to use the 
    shapes layer to add ROIs.   
    
    If the user is interesed in adding different ROI's (e.g. Tumor, Stroma, Tumor-Stromal-interface), 
    each of these should be drawn in a seperate `shape layer`. 
    
    Please note that a single `shape layer` can contain multiple seperate annotated regions. 
    All annotated regions within a single shape layer will be pooled into a single ROI. 
    
    The `shape layers` can be **renamed** to users choice of ROI name.  
    
    Please note: All ROI's have to be unique and cannot overlap.  
    
    *This function could also be used as a QC step to annotate regions of poor/ good quality 
    and later subsetted (keep or remove) for further analysis.*  
    
## Function
"""


# Library
#%gui qt
try:
    import napari
    from napari.layers import Shapes
    from napari.utils.notifications import show_info
except:
    pass
import pandas as pd
import random
import tifffile as tiff

import dask.array as da
from dask.cache import Cache
import zarr
import os
import matplotlib.patches as mpatches
import numpy as np
import scipy.spatial.distance as sdistance
from joblib import Parallel, delayed



cache = Cache(2e9)  # Leverage two gigabytes of memory
cache.register()


# Function
def addROI_image (image_path, adata, subset=None,imageid='imageid', overlay=None, flip_y=True,
                    overlay_category=None,markers=None,channel_names='default',
                    x_coordinate='X_centroid',y_coordinate='Y_centroid',point_size=10,
                    point_color=None,seg_mask=None,
                    n_jobs=-1, verbose=False, 
                    overwrite=True, label='ROI', **kwargs):
    """
Parameters:
    image_path : string  
        Location to the image file (TIFF, OME.TIFF, ZARR supported)  

    adata : AnnData Object  
    
    subset : list, optional  
        Name of the image to which the ROI is to be added. Note if you have multiple images in the 
        adata object, you will need to add ROI's to each image one after the other independently.  
        
    imageid : string, optional  
        In the event that the adata object contains multiple images, it is
        important that ROIs are added to each image seperately. Pass the name 
        of the column that contains the `imageid` and use it in conjunction with
        the `subset` parameter to add ROI's to a specific image.

    seg_mask: string  
        Location to the segmentation mask file.  

    overlay : string, optional  
        Name of the column with any categorical data such as phenotypes or clusters.
        
    flip_y : bool, optional  
        Flip the Y-axis if needed. Some algorithms output the XY with the Y-coordinates flipped.
        If the image overlays do not align to the cells, try again by setting this to `False`.
        
    overlay_category : list, optional  
        If only specfic categories within the overlay column is needed, pass their names as a list.
        If None, all categories will be used.

    markers : list, optional  
        Markers to be included. If none, all markers will be displayed.

    channel_names : list, optional  
        List of channels in the image in the exact order as image. The default is `adata.uns['all_markers']`

    x_coordinate : string, optional  
        X axis coordinate column name in AnnData object.

    y_coordinate : string, optional  
        Y axis coordinate column name in AnnData object.

    point_size : int, optional  
        point size in the napari plot.
    
    overwrite : bool, optional  
        In the event you have multiple images in the adata object, ROI can be added to each image
        independently using the `imageid` and `subset` parameter. If you wish the results to be
        all saved with in the same column set this parameter to `False`. By default, the 
        function will overwrite the `label` column. 
        
    n_jobs : int, optional  
        Number of cores to use. Default is to use all available cores.  
    
    label : string, optional  
        Key for the returned data, stored in `adata.obs`.  

    **kwargs  
        Other arguments that can be passed to napari viewer
    
Returns:
    Modified Anndata object.

Example:
```python
    image_path = '/Users/aj/Desktop/ptcl_tma/image.ome.tif'
    adata = sm.pl.addROI_image (image_path, adata, label="Tumor Regions")
```
    """
    
    #TODO
    # - ADD Subset markers for ZARR ssection
    # - Ability to use ZARR metadata if available
    
    # create data matrix that has the co-ordinates
    data = pd.DataFrame(adata.obs)[[x_coordinate, y_coordinate, imageid]]

    # subset the data if needed
    if subset is not None:
        # convert string to list
        if isinstance(subset, str): 
            subset = [subset]
        # subset data
        sub_data = data[data['imageid'].isin(subset)]
    else:
        sub_data = data
        
    
    # adding option to load just the image without an adata object
    if adata is None:
        channel_names = None
    else: 
        # All operations on the AnnData object is performed first
        # Plot only the Image that is requested
        if subset is not None:
            adata_subset = adata[adata.obs[imageid].isin(subset)]
        else:
            adata_subset = adata.copy()
    
        # Recover the channel names from adata
        if channel_names == 'default':
            channel_names = adata_subset.uns['all_markers']
        else:
            channel_names = channel_names
    
        # Index of the marker of interest and corresponding names
        if markers is None:
            idx = list(range(len(channel_names)))
            channel_names = channel_names
        else:
            idx = []
            for i in markers:
                idx.append(list(channel_names).index(i))
            channel_names = markers
        
        
        # Load the segmentation mask
        if seg_mask is not None:
            seg_m = tiff.imread(seg_mask)
            if (len(seg_m.shape) > 2) and (seg_m.shape[0] > 1):
                seg_m = seg_m[0]    
 
    # Operations on the OME TIFF image is performed next
    # check the format of image
    if os.path.isfile(image_path) is True:  
        image = tiff.TiffFile(image_path, is_ome=False) #is_ome=False
        z = zarr.open(image.aszarr(), mode='r') # convert image to Zarr array
        # Identify the number of pyramids
        n_levels = len(image.series[0].levels) # pyramid
        
        # If and if not pyramids are available
        if n_levels > 1:
            pyramid = [da.from_zarr(z[i]) for i in range(n_levels)]
            multiscale = True
        else:
            pyramid = da.from_zarr(z)
            multiscale = False  
   
        # subset channels of interest
        if markers is not None:
            if n_levels > 1:
                for i in range(n_levels-1):
                    pyramid[i] = pyramid[i][idx, :, :]
                n_channels = pyramid[0].shape[0] # identify the number of channels
            else:
                pyramid = pyramid[idx, :, :]
                n_channels = pyramid.shape[0] # identify the number of channels
        else:
            if n_levels > 1:
                n_channels = pyramid[0].shape[0]
            else:
                n_channels = pyramid.shape[0]
            
        # check if channel names have been passed to all channels
        if channel_names is not None:
            assert n_channels == len(channel_names), (
                f'number of channel names ({len(channel_names)}) must '
                f'match number of channels ({n_channels})'
            )

        # Load the viewer
        viewer = napari.view_image(
            pyramid, multiscale=multiscale, channel_axis=0,
            visible=False, 
            name = None if channel_names is None else channel_names,
            **kwargs
        )
         
    # Operations on the ZARR image
    # check the format of image
    if os.path.isfile(image_path) is False: 
        #print(image_path)
        viewer = napari.Viewer()
        viewer.open(image_path, multiscale=True,
                    visible=False,
                    name = None if channel_names is None else channel_names)

    # Add the seg mask
    if seg_mask is not None:
        viewer.add_labels(seg_m, name='segmentation mask', visible=False)

    # Add phenotype layer function
    def add_phenotype_layer (adata, overlay, phenotype_layer,x,y,viewer,point_size,point_color):
        coordinates = adata[adata.obs[overlay] == phenotype_layer]
        # Flip Y AXIS if needed
        if flip_y is True:
            coordinates = pd.DataFrame({'y': coordinates.obs[y],'x': coordinates.obs[x]})
        else:  
            coordinates = pd.DataFrame({'x': coordinates.obs[x],'y': coordinates.obs[y]})

        #points = coordinates.values.tolist()
        points = coordinates.values
        if point_color is None:
            r = lambda: random.randint(0,255) # random color generator
            point_color = '#%02X%02X%02X' % (r(),r(),r()) # random color generator
        viewer.add_points(points, size=point_size,face_color=point_color,visible=False,name=phenotype_layer)

    if overlay is not None:
        # categories under investigation
        if overlay_category is None:
            available_phenotypes = list(adata_subset.obs[overlay].unique())
        else:
            available_phenotypes = overlay_category

        # Run the function on all phenotypes
        for i in available_phenotypes:
            add_phenotype_layer (adata=adata_subset, overlay=overlay,
                                    phenotype_layer=i, x=x_coordinate, y=y_coordinate, viewer=viewer,
                                    point_size=point_size,point_color=point_color)
     
    # Intiate an ROI layer
    shape_layer = viewer.add_shapes(name=label)
    shape_layer.mode = 'add_polygon'
    #_ = show_info('Draw ROIs')
    
    
    # helper functions
    def ellipse_points_to_patch(vertex_1, vertex_2,co_vertex_1, co_vertex_2):
        
        """
        Parameters
        ----------
        vertex_1, vertex_2, co_vertex_1, co_vertex_2: array like, in the form of (x-coordinate, y-coordinate)
        """
        v_and_co_v = np.array([
            vertex_1, vertex_2,
            co_vertex_1, co_vertex_2
        ])
        centers = v_and_co_v.mean(axis=0)
    
        d = sdistance.cdist(v_and_co_v, v_and_co_v, metric='euclidean')
        width = d[0, 1]
        height = d[2, 3]
    
        vector_2 = v_and_co_v[1] - v_and_co_v[0]
        vector_2 /= np.linalg.norm(vector_2)
    
        angle = np.degrees(np.arccos([1, 0] @ vector_2))
    
        ellipse_patch = mpatches.Ellipse(
            centers, width=width, height=height, angle=angle
        )
        return ellipse_patch
    
    # block the viewer until ROI is added
    a = """
        Opening Napari;
        Add shape layers (on left) to draw ROI's. 
        Rename the shape layer to give a name to your ROI
        Multiple shape layers are supported
        ROI's should not overlap
        Close Napari to save ROI's.
        """
    print(a)
    viewer.show(block=True)    
    

    # Find all the shape layers
    my_shapes = [layer for layer in viewer.layers if isinstance(layer, Shapes)]
    # loop through the layers to find their names
    shape_names = []
    added_rois = []
    for i in my_shapes:
        shape_names.append(i.name)
        added_rois.append(len(viewer.layers[i.name].data))
        
    if any(y > 0 for y in added_rois):
        # Loop through all the Shape layers and extract the vertices and shape type
        all_rois = pd.DataFrame()
        for i in shape_names:
            # return shape vertices
            ver = viewer.layers[i].data
            # return shape shape
            structure = viewer.layers[i].shape_type
            # Each layer may contain multiple ROI's with different shapes (handle that)
            napari_roi_table = pd.DataFrame(dict(vertices=[np.fliplr(v) for v in ver], type=[str(s) for s in structure], ROI=i))
                
            # convert gathered ROIs into mpatches
            for i in range(len(napari_roi_table.index)):
                # ellipse
                if napari_roi_table['type'][i] == 'ellipse':
                    napari_roi_table.loc[i:, 'mpatch'] = ellipse_points_to_patch(napari_roi_table['vertices'][i][0],
                                                                                napari_roi_table['vertices'][i][1],
                                                                                napari_roi_table['vertices'][i][2],
                                                                                napari_roi_table['vertices'][i][3])
                # polygon, rectangle, line
                elif napari_roi_table['type'][i] in ['rectangle', 'polygon', 'path']:
                    napari_roi_table.loc[i:, 'mpatch'] = mpatches.Polygon(napari_roi_table['vertices'][i], closed=True)
                else:
                    raise ValueError
    
            # merge the final ROI's across all shape layers
            all_rois = pd.concat ([all_rois, napari_roi_table])
            all_rois['id'] = range(all_rois.shape[0])
            all_rois.index = all_rois['id']
            
        # Find cells within ROI and add it to the dataframe
        def add_roi_internal (roi_id):
            roi_mpatch = all_rois[all_rois['id'] == roi_id]['mpatch'][roi_id]
            inside = sub_data[roi_mpatch.contains_points(sub_data[[x_coordinate, y_coordinate]])]
            inside['ROI_internal'] = all_rois[all_rois['id'] == roi_id]['ROI'][roi_id]
            # return
            return inside
        
        print("Identifying cells within selected ROI's")
        # all ROI cells
        roi_list = all_rois['id'].unique()
        #final_roi = list()
        #for i in roi_list: 
        #    roi_mpatch = all_rois[all_rois['id'] == i]['mpatch'][i]
        #    inside = sub_data[roi_mpatch.contains_points(sub_data[[x_coordinate, y_coordinate]])]
        #    inside['ROI_internal'] = all_rois[all_rois['id'] == i]['ROI'][i]
        #    final_roi.append(inside)
        
        final_roi = Parallel(n_jobs=n_jobs, verbose=verbose)(delayed(add_roi_internal)(roi_id=i) for i in roi_list)  
        
        # Merge all into a single DF
        final_roi = pd.concat(final_roi)[['ROI_internal']]
        
        # Add the list to obs
        result = pd.merge(data, final_roi, left_index=True, right_index=True, how='outer')
        
        # Reindex
        result = result.reindex(adata.obs.index)
        
        # check if adata already has a column with the supplied label
        # if avaialble overwrite or append depending on users choice
        if label in adata.obs.columns:
            if overwrite is False:
                # Append
                # retreive the ROI information
                old_roi = adata.obs[label]
                combined_roi = pd.merge(result, old_roi, left_index=True, right_index=True, how='outer')
                combined_roi['ROI_internal'] = combined_roi['ROI_internal'].fillna(combined_roi[label])
            else:
                # Over write
                combined_roi = result.copy()
                combined_roi['ROI_internal'] = combined_roi['ROI_internal'].fillna('Other')     
        else:
            # if label is not present just create a new one
            combined_roi = result.copy()
            combined_roi['ROI_internal'] = combined_roi['ROI_internal'].fillna('Other')     
        
        # Add to adata
        adata.obs[label] = combined_roi['ROI_internal']    
        
        # return
        return adata






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# # Created on Mon Dec 14 17:45:14 2020
# @author: Ajit Johnson Nirmal 
"""
!!! abstract "Short Description"
    `sm.tl.spatial_cluster`: This function allows users to cluster the spatial neighbourhood matrix 
    genereated by either `sm.tl.spatial_expression`, `sm.tl.spatial_count`, `sm.tl.spatial_lda` etc.

## Function
"""

# Import library
from scimap.tools._cluster import cluster
import anndata as ad
import argparse
import sys
import pathlib

def main(argv=sys.argv):
    parser = argparse.ArgumentParser(
        description='This function allows users to cluster the spatial neighbourhood matrix genereated by either `sm.tl.spatial_expression` or `sm.tl.spatial_count` methods.'
    )
    parser.add_argument(
        '--adata', required=True, 
        help='AnnData object loaded into memory or path to AnnData object.'
    )
    parser.add_argument(
        '--df_name', type=str, required=False, default='spatial_count',
        help='Label of the spatial analysis performed. By default if `sm.tl.spatial_count` was run the results will be saved under `spatial_count` and if `sm.tl.spatial_expression` was run, the results will be saved under `spatial_expression`.'
    )
    parser.add_argument(
        '--method', type=str, required=False, default='kmeans',
        help='Clustering method to be used- Implemented methods- kmeans, phenograph, leiden and parc.'
    )
    parser.add_argument(
        '--k', type=int, required=False, default=10,
        help='Number of clusters to return when using K-Means clustering.'
    )
    parser.add_argument(
        '--n_pcs', type=int, required=False, default=None,
        help='Number of PCs to be used in leiden clustering. By default it uses all PCs.'
    )
    parser.add_argument(
        '--resolution', type=float, required=False, default=1,
        help='A parameter value controlling the coarseness of the clustering. Higher values lead to more clusters.'
    )
    parser.add_argument(
        '--phenograph_clustering_metric', type=str, required=False, default='euclidean',
        help='Distance metric to define nearest neighbors. Note that performance will be slower for correlation and cosine. Available methods- cityblock’, ‘cosine’, ‘euclidean’, ‘manhattan’, braycurtis’, ‘canberra’, ‘chebyshev’,  ‘correlation’, ‘dice’, ‘hamming’, ‘jaccard’, ‘kulsinski’, ‘mahalanobis’, ‘minkowski’, ‘rogerstanimoto’, ‘russellrao’, ‘seuclidean’, ‘sokalmichener’, ‘sokalsneath’, ‘sqeuclidean’, ‘yule’'
    )
    parser.add_argument(
        '--nearest_neighbors', type=int, required=False, default=30,
        help='Number of nearest neighbors to use in first step of graph construction. This parameter is used both in leiden and phenograph clustering.'
    )
    parser.add_argument(
        '--random_state', type=int, required=False, default=0,
        help='Change the initialization of the optimization.'
    )
    parser.add_argument(
        '--label', type=str, required=False, default=None,
        help='Key or optional column name for the returned data, stored in `adata.obs`. The default is adata.obs [method used].'
    )
    parser.add_argument(
        '--output_dir', type=str, required=False, default=None,
        help='Path to output directory.'
    )
    args = parser.parse_args(argv[1:])
    print(vars(args))
    spatial_cluster(**vars(args))


# Function
def spatial_cluster (adata, df_name='spatial_count', method = 'kmeans',k=10,
                     n_pcs=None, resolution=1, phenograph_clustering_metric='euclidean', 
                     nearest_neighbors=30, random_state=0,label=None, output_dir=None):
    """
    

Parameters:
    adata : AnnData object loaded into memory or path to AnnData object.
    
    df_name : string, required  
        Label of the spatial analysis performed.
        By default if `sm.tl.spatial_count` was run the results will be saved under `spatial_count` and
        if `sm.tl.spatial_expression` was run, the results will be saved under `spatial_expression`.

    method : string, optional  
        Clustering method to be used- Implemented methods- kmeans, phenograph and leiden.

    k : int, optional  
        Number of clusters to return when using K-Means clustering.

    phenotype : string, optional  
        The column name that contains the cluster/phenotype information.

    n_pcs : int, optional  
        Number of PC's to be used in leiden clustering. By default it uses all PC's.

    resolution : float, optional  
        A parameter value controlling the coarseness of the clustering. 
        Higher values lead to more clusters.

    phenograph_clustering_metric : string, optional  
        Distance metric to define nearest neighbors. Note that performance will be slower for correlation and cosine. 
        Available methods- cityblock’, ‘cosine’, ‘euclidean’, ‘manhattan’, braycurtis’, ‘canberra’, ‘chebyshev’, 
        ‘correlation’, ‘dice’, ‘hamming’, ‘jaccard’, ‘kulsinski’, ‘mahalanobis’, ‘minkowski’, ‘rogerstanimoto’, 
        ‘russellrao’, ‘seuclidean’, ‘sokalmichener’, ‘sokalsneath’, ‘sqeuclidean’, ‘yule’

    nearest_neighbors : int, optional  
        Number of nearest neighbors to use in first step of graph construction. 
        This parameter is used both in leiden and phenograph clustering.

    random_state : int, optional  
        Change the initialization of the optimization.

    label : string, optional  
        Key or optional column name for the returned data, stored in `adata.obs`. The default is adata.obs [spatial_method used].
    
    output_dir : string, optional  
        Path to output directory.

Returns:
    adata : AnnData Object  
        Returns an updated anndata object with a new column. check- adata.obs [spatial_method used]
        
Example:
```python
    adata = sm.tl.spatial_cluster (adata, k= 10, method = 'kmeans') # results will be saved under adata.obs['spatial_kmeans']
```
    """

    # Load the andata object    
    if isinstance(adata, str):
        imid = str(adata.rsplit('/', 1)[-1])
        adata = ad.read(adata)
    else:
        adata = adata
    
    # Make a copy of adata to modify
    adata_copy = adata.copy()
    
    # Error check
    try:
        adata_copy.uns[df_name]
    except KeyError:
        print (str('Supplied df_name not found, please run `sm.tl.spatial_expression` or LDA, counts or other similar methods'))
    
    # Crete a new anndata object with the user defined spatial information
    adata_new = ad.AnnData(adata_copy.uns[df_name].fillna(0))
    adata_new.obs = adata_copy.obs
    
    # Create a meaningful label name
    if label is None:
        label = 'spatial_' + str(method)
    
    # Run the clustering algorithm
    adata_new = cluster (adata = adata_new,
                         method = method,
                         k=k, 
                         n_pcs=n_pcs, 
                         resolution=resolution,
                         phenograph_clustering_metric=phenograph_clustering_metric,
                         nearest_neighbors=nearest_neighbors, 
                         use_raw=False, 
                         random_state=random_state,
                         label=label)
    
    # Get the clusters and append that to original adata object
    result = adata_new.obs[label]
    result = result.reindex(adata.obs.index)
    adata.obs[label] = result
    
    
    # Save data if requested
    if output_dir is not None:
        output_dir = pathlib.Path(output_dir)
        output_dir.mkdir(exist_ok=True, parents=True)
        adata.write(output_dir / imid)
    else:    
        # Return data
        return adata

if __name__ == '__main__':
    main()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#Created on Wed Aug 19 15:00:39 2020
# @author: Ajit Johnson Nirmal

"""
!!! abstract "Short Description"
    `sm.tl.spatial_expression`: The function allows users to compute a neighbourhood weighted matrix 
    based on the expression values.

    The function supports two methods to define a local neighbourhood  
    **Radius method**: Can be used to identifies the neighbours within a user defined radius for every cell.  
    **KNN method**: Can be used to identifies the neighbours based on K nearest neigbours for every cell  

    The resultant proportion matrix is saved with `adata.uns`. 

    This can be further clustered to identify similar neighbourhoods. 
    Use the [spatial_cluster] function to further group the neighbourhoods into 
    Reccurent Cellular Neighbourhoods (RCNs)

## Function
"""

# Import library
import pandas as pd
import numpy as np
from sklearn.neighbors import BallTree
from scipy.sparse import csr_matrix
from scipy.sparse import lil_matrix
import scipy
import argparse
import sys
import anndata
import pathlib


def main(argv=sys.argv):
    parser = argparse.ArgumentParser(
        description='The function allows users to compute a neighbourhood weighted matrix based on the expression values.'
    )
    parser.add_argument(
        '--adata', required=True, 
        help='AnnData object loaded into memory or path to AnnData object.'
    )
    parser.add_argument(
        '--x_coordinate', type=str, required=False, default='X_centroid',
        help='Column name containing the x-coordinates values.'
    )
    parser.add_argument(
        '--y_coordinate', type=str, required=False, default='Y_centroid',
        help='Column name containing the y-coordinates values.'
    )
    parser.add_argument(
        '--method', type=str, required=False, default='radius',
        help='Two options are available: a) `radius`, b) `knn`.'
    )
    parser.add_argument(
        '--radius', type=int, required=False, default=30,
        help='The radius used to define a local neighbhourhood.'
    )
    parser.add_argument(
        '--knn', type=int, required=False, default=10,
        help='Number of cells considered for defining the local neighbhourhood.'
    )
    parser.add_argument(
        '--imageid', type=str, required=False, default='imageid',
        help='Column name of the column containing the image id.'
    )
    parser.add_argument(
        '--use_raw', type=bool, required=False, default=True,
        help='Argument to denote whether to use the raw data or scaled data after applying `sm.pp.rescale`.'
    )
    parser.add_argument(
        '--log', type=bool, required=False, default=True,
        help='If `True`, the log of raw data is used. Set use_raw = `True` for this to take effect.'
    )
    parser.add_argument(
        '--subset', type=str, required=False, default=None,
        help='imageid of a single image to be subsetted for analyis.'
    )
    parser.add_argument(
        '--label', type=str, required=False, default='spatial_expression',
        help='Key for the returned data, stored in `adata.uns`.'
    )
    parser.add_argument(
        '--output_dir', type=str, required=False, default=None,
        help='Path to output directory.'
    )
    args = parser.parse_args(argv[1:])
    print(vars(args))
    spatial_expression(**vars(args))

# Function
def spatial_expression (adata, 
                        x_coordinate='X_centroid',
                        y_coordinate='Y_centroid',
                        method='radius', radius=30, 
                        knn=10, imageid='imageid', 
                        use_raw=True, log=True, subset=None,
                        label='spatial_expression',
                        output_dir=None):
    """
Parameters:
    adata : AnnData object loaded into memory or path to AnnData object.

    x_coordinate : float, required  
        Column name containing the x-coordinates values.
        
    y_coordinate : float, required  
        Column name containing the y-coordinates values.
        
    method : string, optional  
        Two options are available: a) `radius`, b) `knn`.  
        a) `radius` - Identifies the neighbours within a given radius for every cell.  
        b) `knn` - Identifies the K nearest neigbours for every cell.  
        
    radius : int, optional  
        The radius used to define a local neighbhourhood.
        
    knn : int, optional  
        Number of cells considered for defining the local neighbhourhood.
        
    imageid : string, optional  
        Column name of the column containing the image id.
        
    subset : string, optional  
        imageid of a single image to be subsetted for analyis.
        
    use_raw : boolian, optional  
        Argument to denote whether to use the raw data or scaled data after applying `sm.pp.rescale`.

    log : boolian, optional  
        If `True`, the log of raw data is used. Set use_raw = `True` for this to take effect. 
        
    label : string, optional  
        Key for the returned data, stored in `adata.uns`.

    output_dir : string, optional  
        Path to output directory.

Returns:
    adata : AnnData object  
        Updated AnnData object with the results stored in `adata.uns ['spatial_expression']`.
        
        
 Example:
    ```python
    # Running the radius method
    adata = sm.tl.spatial_expression (adata, x_coordinate='X_centroid',
                                      y_coordinate='Y_centroid',
                                      method='radius', radius=30, 
                                      imageid='imageid', 
                                      use_raw=True,subset=None,
                                      label='spatial_expression_radius')
    
    # Running the knn method
    adata = sm.tl.spatial_expression (adata, x_coordinate='X_centroid',
                                      y_coordinate='Y_centroid',
                                      method='knn', knn=10, imageid='imageid', 
                                      use_raw=True,subset=None,
                                      label='spatial_expression_knn')
    ```
    """
    
    # Load the andata object    
    if isinstance(adata, str):
        imid = str(adata.rsplit('/', 1)[-1])
        adata = anndata.read(adata)
    else:
        adata = adata

    
    # Error statements
    if use_raw is False:
        if all(adata.X[0] < 1) is False:
            raise ValueError('Please run `sm.pp.rescale` first if you wish to use `use_raw = False`')
        
     
    def spatial_expression_internal (adata_subset, x_coordinate, y_coordinate,log,
                                     method, radius, knn, imageid, use_raw, subset,label):
         
        # Create a DataFrame with the necessary inforamtion
        data = pd.DataFrame({'x': adata_subset.obs[x_coordinate], 'y': adata_subset.obs[y_coordinate]})
        
        # Identify neighbourhoods based on the method used
        # a) KNN method
        if method == 'knn':
            print("Identifying the " + str(knn) + " nearest neighbours for every cell")
            tree = BallTree(data, leaf_size= 2)
            dist, ind = tree.query(data, k=knn, return_distance= True)

            
        # b) Local radius method
        if method == 'radius':
            print("Identifying neighbours within " + str(radius) + " pixels of every cell")
            kdt = BallTree(data, metric='euclidean')
            ind, dist = kdt.query_radius(data, r=radius, return_distance= True)
            
        # Normalize range (0-1) and account for total number of cells 
        d = scipy.sparse.lil_matrix((len(data), len(data)))
        for row, (columns, values) in enumerate(zip(ind, dist)):
            # Drop self-distance element.
            idx = columns != row
            columns = columns[idx]
            values = values[idx]
            if len(values) == 1:
                values = [1.0]
            elif len(values) > 1:
                # Normalize distances.
                values = (values.max() - values) / (values.max() - values.min())
                values /= values.sum()
            # Assign row to matrix.
            d[row, columns] = values
        
        # convert to csr sparse matrix
        wn_matrix_sparse = d.tocsr()
        
        
        # Calculation of spatial lag
        if use_raw==True:
            if log is True:
                spatial_lag = pd.DataFrame(wn_matrix_sparse * np.log1p(adata_subset.raw.X), columns = adata_subset.var.index, index=adata_subset.obs.index)
            else:
                spatial_lag = pd.DataFrame(wn_matrix_sparse * adata_subset.raw.X, columns = adata_subset.var.index, index=adata_subset.obs.index)
        else:
            spatial_lag = pd.DataFrame(wn_matrix_sparse * adata_subset.X, columns = adata_subset.var.index, index=adata_subset.obs.index)
        
        # return value
        return spatial_lag
    
    # Subset a particular image if needed
    if subset is not None:
        adata_list = [adata[adata.obs[imageid] == subset]]
    else:
        adata_list = [adata[adata.obs[imageid] == i] for i in adata.obs[imageid].unique()]
    
    # Apply function to all images and create a master dataframe
    # Create lamda function 
    r_spatial_expression_internal = lambda x: spatial_expression_internal(adata_subset=x, 
                                                                x_coordinate=x_coordinate, 
                                                                y_coordinate=y_coordinate, 
                                                                method=method, radius=radius, 
                                                                knn=knn, imageid=imageid, 
                                                                use_raw=use_raw, subset=subset,
                                                                log=log,
                                                                label=label) 
    all_data = list(map(r_spatial_expression_internal, adata_list)) # Apply function 
    
    # Merge all the results into a single dataframe    
    result = []
    for i in range(len(all_data)):
        result.append(all_data[i])
    result = pd.concat(result, join='outer')  
    
    # Reindex the cells
    result = result.fillna(0)
    result = result.reindex(adata.obs.index)
    
    # Add to adata
    adata.uns[label] = result
    
    # Save data if requested
    if output_dir is not None:
        output_dir = pathlib.Path(output_dir)
        output_dir.mkdir(exist_ok=True, parents=True)
        adata.write(output_dir / imid)
    else:    
        # Return data
        return adata


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Feb 13 21:43:29 2021
@author: Ajit Johnson Nirmal
Lasso selector to annotate regions of interest to the anndata object
"""

# Library
import matplotlib.pyplot as plt
from matplotlib.widgets import LassoSelector
from matplotlib.path import Path
import numpy as np
import pandas as pd


# Function

def add_roi_scatter (adata, method=None, marker=None, threshold=0.5, raw=False, log=False, subset=None, imageid='imageid',
                     x_coordinate='X_centroid', y_coordinate='Y_centroid',size=5,cmap='gist_heat',
                     lasso_alpha=0.8, lasso_linewidth=3, lasso_color='red',
                     roi_name='selected_roi', label='roi_scatter'):
    
    
    # create a copy of the anndata object
    adata = combined_excluded_RJP.copy()
    bdata = adata.copy()
    
    # Subset the image of interest
    if subset is not None:
        bdata = bdata[bdata.obs[imageid] == subset]
        
    # indentify the index of the marker of interest
    if marker is not None:
        marker_index = bdata.var.index.get_loc(marker)
        
    # generate the dataframe needed with 
    if method is None:
        data = pd.DataFrame({'x': bdata.obs[x_coordinate], 'y': bdata.obs[y_coordinate]})
    else:
        if marker is None:
            raise ValueError('Please choose a marker to be able to use the selected method')
    
    # If a method is passed first create the data frame with expression values
    if method is not None and raw is False:
        data = pd.DataFrame({'x': bdata.obs[x_coordinate], 'y': bdata.obs[y_coordinate], marker: bdata.X[:,marker_index]})
        if log is True:
            data[marker] = np.log1p(data[marker])
    elif method is not None and raw is True:
        data = pd.DataFrame({'x': bdata.obs[x_coordinate], 'y': bdata.obs[y_coordinate], marker: bdata.raw.X[:,marker_index]})
        if log is True:
            data[marker] = np.log1p(data[marker])
            
    # If threshold method is passed convert the marker expression to binary
    if method == 'threshold':
        hue = np.array(data[marker])
        hue = ['red' if x >= threshold else 'grey' for x in hue]
    

    # Scatter plot start
    fig, ax = plt.subplots()

    # plot
    if method == 'threshold':
        collection = ax.scatter(x=data['x'], y=data['y'], c=hue, s=size)
    elif method == 'expression':
        collection = ax.scatter(x=data['x'], y=data['y'], c=data[marker], cmap=cmap, s=size)
    else:
        collection = ax.scatter(x=data['x'], y=data['y'], s=size)
    ax.invert_yaxis()
    plt.xticks([]) ; plt.yticks([]);
    
    
    xys = collection.get_offsets()
    ind = []
    
    def onSelect(verts):
        path = Path(verts)
        ind = np.nonzero(path.contains_points(xys))[0]
        adata.obs.loc[data.iloc[ind].index, label] = roi_name
        print('Added Selected ROI to adata.obs.' + str(label))
    
    lineprops = {'color': lasso_color, 'linewidth': lasso_linewidth, 'alpha': lasso_alpha}
    lsso = LassoSelector(ax=ax, onselect=onSelect, lineprops=lineprops)
    
    
    plt.show()
        
        




